from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys
import numpy as np

PROJECT_ROOT_DIR = Path(__file__).parent.parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase


class ReachTargetTermination(TerminationBase):

    def __init__(self,
        integral_time_length: float=1.,
        mu_tolerance: float=1.,
        v_threshold=10., chi_threshold=1., 
        termination_reward: float=1.,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=False,
            env_config=env_config, 
            my_logger=my_logger
        )

        self.integral_time_length = integral_time_length
        
        self.mu_tolerance = mu_tolerance
        self.v_threshold = v_threshold
        self.chi_threshold = chi_threshold

    def _get_termination(self, goal_v: float, goal_chi: float, state_list: List[namedtuple], ):
        
        if len(state_list) < self.integral_window_length:
            return False, False
        else:
            last_mus = np.array([item.mu for item in state_list[-self.integral_window_length:]])

            if np.all(last_mus < self.mu_tolerance):

                v_flag, chi_flag = False, False
                if sum([abs(goal_v - item.v) for item in state_list[-self.integral_window_length:]]) < self.v_integral_threshold:
                    v_flag = True
                if sum([abs(goal_chi - item.chi) for item in state_list[-self.integral_window_length:]]) < self.chi_integral_threshold:
                    chi_flag = True
                if v_flag and chi_flag:
                    return True, False
                else:
                    return False, False
            else:
                return False, False

    def get_termination(self, state, **kwargs) -> Tuple[bool, bool]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        assert "state_list" in kwargs, "参数中需要包括state_list（list[namedtuple]类型，所有历史观测），再调用get_termination方法"
        assert type(kwargs["state_list"]) is list, "state_list参数的类型应该是list[namedtuple]"

        return self._get_termination(
            goal_v=kwargs["goal_v"], 
            goal_chi=kwargs["goal_chi"],
            state_list=kwargs["state_list"]
        )

    def get_termination_and_reward(self, state, **kwargs) -> Tuple[bool, bool, float]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        assert "state_list" in kwargs, "参数中需要包括state_list（list[namedtuple]类型，所有历史观测），再调用get_termination方法"
        assert type(kwargs["state_list"]) is list, "state_list参数的类型应该是list[namedtuple]"

        terminated, truncated = self._get_termination(
            goal_v=kwargs["goal_v"], 
            goal_chi=kwargs["goal_chi"],
            state_list=kwargs["state_list"]
        )
        reward = self.termination_reward if terminated else 0.
        
        return terminated, truncated, reward

    def reset(self):
        pass
    
    @property
    def integral_window_length(self) -> int:
        """v, mu, chi的积分窗口长度

        Returns:
            _type_: _description_
        """
        return round(self.integral_time_length * self.step_frequence)

    @property
    def v_integral_threshold(self):
        """v的误差积分阈值，当v在最后self.integral_window_length上的积分小于该值时，认为v达到目标值

        Returns:
            _type_: _description_
        """
        return self.v_threshold * self.integral_window_length
    
    
    @property
    def chi_integral_threshold(self):
        """chi的误差积分阈值，当chi在最后self.integral_window_length上的积分小于该值时，认为chi达到目标值

        Returns:
            _type_: _description_
        """
        return self.chi_threshold * self.integral_window_length

    def __str__(self) -> str:
        return "reach_target_termination"


from collections import namedtuple
from pathlib import Path
import sys
from copy import deepcopy
import gc
from typing import Union, Dict, Iterable

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from planes.utils.f16Classes import ControlLaw, PlaneModel

class F16Plane(object):

    def __init__(self, 
        env_config: dict
    ) -> None:
        
        self.config = env_config

        self.step_frequence = self.config["task"].get("step_frequence")
        self.h0, self.v0 = self.config["task"].get("h0"), self.config["task"].get("v0")
        self.max_simulate_time = self.config["task"].get("max_simulate_time")
        self.control_mode = self.config["task"].get("control_mode", "guidance_law_mode")  # 两个值：guidance_law_mode，end_to_end_mode
        self.control_mode_guidance_law: bool = (self.control_mode == "guidance_law_mode")

        self.f16cl: ControlLaw = None
        self.f16model: PlaneModel = None

        self.state_list = []
        self.action_list = []
        self.step_cnt = 0

    def reset(self):
        if self.control_mode_guidance_law:
            self.f16cl = ControlLaw(stepTime=self.step_time)  # 控制律模型内有积分，所以也要重新初始化
        self.f16model = PlaneModel(self.h0, self.v0, stepTime=self.step_time)  # 飞机初始化，会配平，

        self.step_cnt = 0
        self.current_obs = self.f16model.getPlaneState()
        self.state_list = [self.current_obs]
        self.action_list = []

        return self.state_list[-1]

    def close(self):
        # 先回收旧的self.f16cl和self.f16model，否则会存在内存持续增长的情况！！！！！
        if self.f16cl is not None:
            # if hasattr(self.f16cl, "f16CL"):
            #     del self.f16cl.f16CL
            del self.f16cl
        if self.f16model is not None:
            # if hasattr(self.f16model, "f16Model"):
            #     del self.f16model.f16Model
            del self.f16model
        gc.collect()

    def step(self, action: Union[Dict, Iterable]) -> Dict:
        """

        Args:
            action (Union[Dict, Iterable]): {"p":xx, "nz":xx, "pla":xx} or [xx, xx, xx]

        Returns:
            Dict: {
                'lef': 0.0, 
                'npos': 202.93280838546622, 
                'epos': -0.0003178369761097856, 
                'h': 4997.945252461648, 
                'alpha': -3.8763871428920833, 
                'beta': 0.0028827414834535683, 
                'phi': -0.0007924768497675406, 
                'theta': -6.5392414504888485, 
                'psi': -0.003222236733571302, 
                'p': 0.0012092603223480781, 
                'q': -2.301811791467605, 
                'r': 0.0018719855098707125, 
                'v': 206.42898450413645, 
                'vn': 206.23466760853253, 
                've': -0.0014038370582571266, 
                'vh': -9.591781654339009, 
                'nx': 0.6544365026301127, 
                'ny': -0.00026289226844918224, 
                'nz': -0.8983464478861598, 
                'ele': -3.725121874665151, 
                'ail': -0.0017574644108460835, 
                'rud': -0.003584270151159641, 
                'thrust': 1.0, 
                'lon': 122.42499999666104, 
                'lat': 31.426828065405605, 
                'mu': -2.126962766563186, 
                'chi': -0.00031410986307750424
            }
        """
        self.step_cnt += 1
        self.action_list.append(action)

        if self.control_mode_guidance_law:
            # 使用控制律模型
            if isinstance(action, Iterable):
                guide_output = {
                    "p": action[0],
                    "nz": action[1],
                    "pla": action[2],
                    "rud": 0.
                }  # [p, nz, pla, rud], rud置为0
            elif isinstance(action, Dict):
                guide_output = {
                    "p": action["p"],
                    "nz": action["nz"],
                    "pla": action["pla"],
                    "rud": 0.
                }
            else:
                raise TypeError("plane: action must be of Iterable or Dict!")

            self.f16cl.step(guide_output, self.current_obs)
            control_law_output = self.f16cl.getOutputDict()
            self.f16model.step(control_law_output)
        else:
            # 不用控制律模型
            if isinstance(action, Iterable):
                end2end_action = {
                    "ail": action[0],
                    "ele": action[1],
                    "rud": action[2],
                    "pla": action[3]
                }
            elif isinstance(action, Dict):
                end2end_action = action
            else:
                raise TypeError("plane: action must be of Iterable or Dict!")
            
            self.f16model.step(end2end_action)
        
        next_obs = self.f16model.getPlaneState()

        self.state_list.append(next_obs)
        self.current_obs = deepcopy(next_obs)
        return self.state_list[-1]

    @property
    def step_time(self):
        return 1. / self.step_frequence

    @staticmethod
    def get_action_vars(control_mode: str="guidance_law_mode"):
        if control_mode == "guidance_law_mode":
            return namedtuple("action_vars", ["p", "nz", "pla"])
        elif control_mode == "end_to_end_mode":
            return namedtuple("action_vars", ["ail", "ele", "rud", "pla"])
        else:
            raise TypeError("plane: control_mode must be guidance_law_mode or end_to_end_mode!")
    
    @staticmethod
    def get_action_lower_bounds(control_mode: str="guidance_law_mode"):
        action_vars_type = F16Plane.get_action_vars(control_mode)
        
        if control_mode == "guidance_law_mode":
            return action_vars_type(p=-180., nz=-4., pla=0.)
        elif control_mode == "end_to_end_mode":
            return action_vars_type(ail=-21.5, ele=-25., rud=-30, pla=0.)
        else:
            raise TypeError("plane: control_mode must be guidance_law_mode or end_to_end_mode!")

    @staticmethod
    def get_action_higher_bounds(control_mode: str="guidance_law_mode"):
        action_vars_type = F16Plane.get_action_vars(control_mode)

        if control_mode == "guidance_law_mode":
            return action_vars_type(p=180., nz=9., pla=1.)
        elif control_mode == "end_to_end_mode":
            return action_vars_type(ail=21.5, ele=25., rud=30, pla=1.)
        else:
            raise TypeError("plane: control_mode must be guidance_law_mode or end_to_end_mode!")



import os
from gymnasium.envs.registration import register

with open(os.path.join(os.path.dirname(__file__), "version.txt"), "r") as file_handler:
    __version__ = file_handler.read().strip()

register(
    id="FlyCraft-v0",
    entry_point="flycraft.env:FlyCraftEnv",
)




from math import sqrt

R = 287.05
T0 = 288.15
gamma = 1.4

def temperature(altt=4000):
    temp = T0 - 0.0065 * altt
    if altt >= 11000.0:
       temp = 216.65
    return temp


def airSpeed(altt=0):
    return sqrt(gamma * R * temperature(altt))

def TAS2Mach(TAS, altt=4000):
    mach = TAS / airSpeed(altt)
    return mach
    
def Mach2TAS(mach, altt=4000):
    TAS = mach * airSpeed(altt)
    return TAS


import os
import sys
from pathlib import Path
from math import radians, sin, cos, tan, atan2, asin
from ctypes import CDLL, c_char_p, c_double, c_void_p, c_int

PROJECT_ROOT_DIR = Path(__file__).parent.parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from planes.utils.f16_utils import currentOps, currentArchi, pythonVersion
from planes.utils.f16_utils.util import Mach2TAS


def checkPlatform():
    if currentOps == 'Windows':
        suffix = '.dll'
        if currentArchi == '32':
            end = 'x86'
        else:
            end = 'x64'
    else:
        suffix = '.so'
        end = 'ubuntu'
    path = 'libs/plane/f16/' + end
    return path, suffix

class Guide(object):
    def __init__(self, planeType=None):
        self.planeType = planeType
        self.initDll()
        self.initLog()
    
    def __del__(self):
        self.log.close()
        
    def initDll(self):
        path, ends = checkPlatform()
        # dllGuide = os.path.join(path, "Guide1" + ends)
        dllGuide = os.path.join(PROJECT_ROOT_DIR, "planes", "utils", path, "Guide1" + ends)
        self.f16Guide = CDLL(dllGuide)
        self.initArgs()
    
    def initLog(self):
        # self.log = open('logs/guideInOut.csv', 'w')
        self.log = open(os.path.join(PROJECT_ROOT_DIR, "planes", "utils", "logs", "guideInOut.csv"), 'w')
        usefulKeys = ['cmdMu', 'mu', 'dchi', 'phi', 'cmdPhi', 'cmdNz']
        self.log.write(','.join(usefulKeys))
        self.log.write('\n')
    
    def initArgs(self):
        c_double_array4 = c_double*4
        c_double_array19 = c_double*19
        self.inputGuide = c_double_array19()
        self.outputGuide = c_double_array4()
        self.outputDict = {}
    
    def step(self, cmdDict, planeStateDict):
        self.setInput(cmdDict, planeStateDict)
        self.f16Guide.Guide_OneStep(self.inputGuide)
        self.f16Guide.Guide_GetOutput(self.outputGuide)
    
    def setWCmd(self, cmdDict):
        self.inputGuide[0] = cmdDict['v']       # Cmd_V, m/s
        self.inputGuide[1] = cmdDict['mu']         # Cmd_Vtheta, deg
        self.inputGuide[2] = cmdDict['chi']        # Cmd_Vpsi, deg
        self.inputGuide[3] = 0         # K_switch, proportion of LV cmd, 0 is only W, 1 is only LV
        self.inputGuide[4] = 1         # Nz_c_LV
        self.inputGuide[5] = 0         # Phi_c_LV
        self.inputGuide[6] = 0         # Thr_c_LV
        self.inputGuide[7] = 1         # Mode, 1W, 2LV, 3Best turn
        self.log.write(str(cmdDict['mu']))
    
    def setPlaneState(self, planeStateDict):
        self.inputGuide[8] = planeStateDict['psi']          #psi_p
        self.inputGuide[9] = planeStateDict['theta']        #pit_p
        self.inputGuide[10] = planeStateDict['phi']         #bank_p
        self.inputGuide[11] = planeStateDict['v']       #VIAS_p
        self.inputGuide[12] = planeStateDict['vh']      #Vh 
        self.inputGuide[13] = planeStateDict['vn']      #Vn Vx
        self.inputGuide[14] = planeStateDict['ve']      #Ve Vy
        self.inputGuide[15] = planeStateDict['p']       #p
        self.inputGuide[16] = planeStateDict['h']       #h
        self.inputGuide[17] = 0                  #cmdP
        self.inputGuide[18] = 0                  #cmdRudder

    def setInput(self, cmdDict, planeStateDict):
        self.setWCmd(cmdDict)
        self.setPlaneState(planeStateDict)
    
    def getOutputDict(self):
        self.outputDict['p'] = self.outputGuide[0]      #roll rate: -180~180 deg/s
        self.outputDict['nz'] = self.outputGuide[1]     #g load: -3~8
        self.outputDict['pla'] = self.outputGuide[2]    #throttle: 0~1.0
        self.outputDict['rud'] = self.outputGuide[3]    #RUD
        return self.outputDict

# stepTime = 0.01

class ControlLaw(object):
    def __init__(self, type='f16', stepTime=0.01):
        self.stepTime = stepTime
        self.initDll()
        self.initArgs()
        self.initDllFuncTypes()
        self.initControlLaw()
    
    def __del__(self):
        self.f16CL.clExit(self.clPtr)
    
    def initDll(self):
        path, ends = checkPlatform()
        # dllCL = os.path.join(path, "f16cl" + ends)
        dllCL = os.path.join(PROJECT_ROOT_DIR, "planes", "utils", path, "f16cl" + ends)
        self.f16CL = CDLL(dllCL)
    
    def initArgs(self):
        c_double_array4 = c_double*4
        c_double_array10 = c_double*10
        self.inState = c_double_array10()
        self.inCmd = c_double_array4()
        self.output = c_double_array4()
        self.outputDict = {}
    
    def initDllFuncTypes(self):
        self.f16CL.init.argtypes = [c_double, c_char_p]
        self.f16CL.init.restype = c_void_p
        self.f16CL.step.argtypes = [c_void_p, c_double*4, c_double*10, c_double*4]
        self.f16CL.setLog.argtypes = [c_void_p, c_char_p]
    
    def initControlLaw(self):
        clParamDir = os.path.join(PROJECT_ROOT_DIR, "planes", "utils", "f16", "lib", "cl")
        clParamDir = bytes(clParamDir, 'utf-8')
        self.clPtr = self.f16CL.init(self.stepTime, clParamDir)
    
    def setInState(self, planeStateDict):
        self.inState[0] = planeStateDict['h']
        self.inState[1] = planeStateDict['v']
        self.inState[2] = planeStateDict['alpha']
        self.inState[3] = planeStateDict['beta']
        self.inState[4] = planeStateDict['phi']
        self.inState[5] = planeStateDict['theta']
        self.inState[6] = planeStateDict['p']
        self.inState[7] = planeStateDict['q']
        self.inState[8] = planeStateDict['r']
        self.inState[9] = planeStateDict['nz']
    
    def setInCmd(self, cmdDict):
        self.inCmd[0] = cmdDict['p']      # roll rate: -85~85 deg/s
        self.inCmd[1] = cmdDict['nz']     # g load: -3~8
        self.inCmd[2] = cmdDict['rud']
        self.inCmd[3] = cmdDict['pla']    # throttle: 0~1.0
    
    def step(self, cmdDict, planeStateDict):
        self.setInCmd(cmdDict)
        self.setInState(planeStateDict)
        self.f16CL.step(self.clPtr, self.inCmd, self.inState, self.output)
    
    def getOutputDict(self):
        self.outputDict['ail'] = self.output[0]
        self.outputDict['ele'] = self.output[1]
        self.outputDict['rud'] = self.output[2]
        self.outputDict['pla'] = self.output[3]
        return self.outputDict
    
    def setLog(self, logName):
        logName = bytes(logName, 'utf-8')
        self.f16CL.setLog(self.clPtr, logName)

class PlaneModel(object):
    def __init__(self, h0, v0, stepTime=0.01):
        self.stepTime = stepTime
        self.initDll()
        self.initArgs()
        self.initDllFuncTypes()
        self.initModel(h0, v0)
    
    def __del__(self):
        pass
        # 下面一行原本是注释状态，取消注释后，程序无法正常运行！！！！！！！！！！！！！！
        # self.f16Model.deletePtr(self.planePtr)
    
    def initDll(self):
        path, ends = checkPlatform()
        # dllModel = os.path.join(path, "f16model" + ends)
        dllModel = os.path.join(PROJECT_ROOT_DIR, "planes", "utils", path, "f16model" + ends)
        self.f16Model = CDLL(dllModel)
    
    def initArgs(self):
        c_double_array4 = c_double*4
        c_double_array31 = c_double*31
        self.inputAct = c_double_array4()
        self.output = c_double_array31()
        self.stateDict = {}
        self.stateKeys = ['lef']
        self.stateKeys += ['npos', 'epos', 'h']
        self.stateKeys += ['alpha', 'beta']
        self.stateKeys += ['phi', 'theta', 'psi']
        self.stateKeys += ['p', 'q', 'r']
        self.stateKeys += ['v', 'vn', 've', 'vh']
        self.stateKeys += ['nx', 'ny', 'nz']
        self.stateKeys += ['ele', 'ail', 'rud', 'thrust']
        self.stateKeys += ['ele', 'ail', 'rud', 'thrust']
        self.stateKeys += ['lon', 'lat']
        self.stateKeys += ['mu', 'chi']
    
    def initDllFuncTypes(self):
        self.f16Model.init.argtypes = [c_double, c_double, c_double, c_int, c_char_p]
        self.f16Model.init.restype = c_void_p
        self.f16Model.setlla.argtypes = [c_void_p, c_double, c_double]
        self.f16Model.setPos.argtypes = [c_void_p, c_double, c_double]
        self.f16Model.setPhi.argtypes = [c_void_p, c_double]
        self.f16Model.setAttitude.argtypes = [c_void_p, c_double, c_double, c_double]
        self.f16Model.stepsim.argtypes = [c_void_p, c_double, c_double, c_double, c_double, c_double*31]
        self.f16Model.getStates.argtypes = [c_void_p, c_double*31]
        self.f16Model.getMaxThrust.argtypes = [c_void_p, c_double, c_double]
        self.f16Model.getMaxThrust.restype = c_double
        self.f16Model.saveStates.argtypes = [c_void_p]
        self.f16Model.loadStates.argtypes = [c_void_p]
        self.f16Model.loadPlaneInfo.argtypes = [c_void_p, c_double*14, c_double*4]
    
    def initModel(self, h0, v0):
        ''' init Plane model '''
        if v0 < 10:
            v0 = Mach2TAS(v0, h0)
        self.h0 = h0
        self.v0 = v0
        self.lon = 122.425
        self.lat = 31.425
        
        self.planeAirLib = os.path.join(PROJECT_ROOT_DIR, "planes", "utils", "f16", "lib", "f16_air.lib")
        libsPath = c_char_p(bytes(self.planeAirLib, 'utf-8'))
        self.planePtr = self.f16Model.init(self.h0, self.v0, self.stepTime, 1, libsPath)
        self.setLonLat(self.lon, self.lat)
    
    def setLonLat(self, lon, lat):
        self.f16Model.setlla(self.planePtr, lon, lat)
        self.f16Model.getStates(self.planePtr, self.output)
    
    def setPlaneAttitude(self, phi, theta, psi):
        self.f16Model.setAttitude(self.planePtr, phi, theta, psi)
        self.f16Model.getStates(self.planePtr, self.output)
    
    def setLog(self, logName):
        logName = bytes(logName, 'utf-8')
        self.f16Model.setLog(self.planePtr, logName)
    
    def setInAct(self, actDict):
        self.inputAct[0] = actDict['ail']
        self.inputAct[1] = actDict['ele']
        self.inputAct[2] = actDict['rud']
        self.inputAct[3] = actDict['pla']
    
    def step(self, actDict):
        self.setInAct(actDict)
        self.f16Model.stepsim(self.planePtr, self.inputAct[0], self.inputAct[1], self.inputAct[2], self.inputAct[3], self.output)

    def getPlaneState(self):
        for index, key in enumerate(self.stateKeys):
            self.stateDict[key] = self.output[index]
        return self.stateDict



import os
import sys
from pathlib import Path
from math import radians, sin, cos

PROJECT_ROOT_DIR = Path(__file__).parent.parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from planes.utils.f16Classes import Guide, ControlLaw, PlaneModel


def testGuide():
    gtmp = Guide()
    cmds = {}
    cmds['v'] = 180
    cmds['mu'] = 0
    cmds['chi'] = 30
    
    stateDict = {}
    stateDict['phi'] = 0
    stateDict['theta'] = 10
    stateDict['psi'] = 0
    stateDict['v'] = 160
    stateDict['mu'] = 0
    stateDict['chi'] = 0
    stateDict['vn'] = stateDict['v'] * cos(radians(stateDict['mu'])) * cos(radians(stateDict['chi']))
    stateDict['ve'] = stateDict['v'] * cos(radians(stateDict['mu'])) * sin(radians(stateDict['chi']))
    stateDict['vh'] = stateDict['v'] * sin(radians(stateDict['mu']))
    stateDict['p'] = 0
    stateDict['h'] = 1000
    for i in range(10):
        gtmp.step(cmds, stateDict)

        #################################
        acts = gtmp.getOutputDict()
        for key in acts:
            print(key, acts[key], end=',')
        print()
        #################################

        for i in range(4):
            print(gtmp.outputGuide[i], end = ',')
        print()

def testCL():
    f16cl = ControlLaw()
    cmds = {}
    cmds['p'] = 180
    cmds['nz'] = 1
    cmds['rud'] = 0
    cmds['pla'] = 1
    
    stateDict = {}
    stateDict['v'] = 160
    stateDict['h'] = 1000
    stateDict['alpha'] = 10
    stateDict['beta'] = 0
    stateDict['phi'] = 0
    stateDict['theta'] = 10
    stateDict['p'] = 0
    stateDict['q'] = 0
    stateDict['r'] = 0
    stateDict['nz'] = 1
    f16cl.step(cmds, stateDict)
    acts = f16cl.getOutputDict()
    for key in acts:
        print(key, acts[key], end = ',')
    print()

def testPlane():
    h0 = 1000
    v0 = 200
    f16model = PlaneModel(h0, v0)
    stsDict = f16model.getPlaneState()  
    # stsDict: lef, npos, epos, h, alpha, beta, phi, theta, psi, p, q, r, v, vn, ve, vh, nx, ny, nz, ele, ail, rud, thrust, lon, lat, mu, chi
    
    # for k in stsDict.keys():
    #     print(k, end=', ')
    # print()
    showKeys = ['p', 'q', 'r']
    for key in showKeys:
        print(stsDict[key], end = ',')
    print()
    
    cmds = {}
    cmds['ail'] = 20
    cmds['ele'] = -1
    cmds['rud'] = 0
    cmds['pla'] = 1
    for i in range(10):
        f16model.step(cmds)
        stsDict = f16model.getPlaneState()
        for key in showKeys:
            print(stsDict[key], end = ',')
        print()

def testFullStep():
    h0 = 1000
    v0 = 200
    gtmp = Guide()
    f16cl = ControlLaw()
    f16model = PlaneModel(h0, v0)
    # f16model.setLog('utils/f16trace4.csv')  # 设置日志名
    stsDict = f16model.getPlaneState()
    
    wCmds = {}
    # wCmds['v'] = 180
    # wCmds['mu'] = 20
    # wCmds['chi'] = 30
    wCmds['v'] = 180
    wCmds['mu'] = -90
    wCmds['chi'] = 0
    for i in range(12000):
        gtmp.step(wCmds, stsDict)
        gout = gtmp.getOutputDict()
        f16cl.step(gout, stsDict)
        clout = f16cl.getOutputDict()
        f16model.step(clout)
        stsDict = f16model.getPlaneState()

if __name__ == '__main__':
    # testGuide()
    # testCL()
    # testPlane()
    testFullStep()


import platform
import os
import sys

currentOps = ""
currentArchi = ""
pythonVersion = sys.version_info.major
esPath = "/home/pi/champ/pythonChamp/Strategy/integrator"

if platform.system() == "Windows":
    currentOps = "Windows"
elif os.path.exists(esPath):
    currentOps = "Rasberrypi"
else:
    currentOps = "Ubuntu"

if platform.architecture()[0] == "64bit":
    currentArchi = "64"
else:
    currentArchi = "32"
    



import sys
from pathlib import Path
from abc import ABC, abstractmethod
import numpy as np
from typing import Dict, Any

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from planes.f16_plane import F16Plane

class Task(ABC):

    def __init__(self, plane: F16Plane) -> None:
        self.plane: F16Plane = plane
        self.goal: np.ndarray = None
    
    @abstractmethod
    def reset(self) -> None:
        """Reset the task: sample a new goal."""

    @abstractmethod
    def get_obs(self) -> np.ndarray:
        """Return the observation associated to the task."""

    @abstractmethod
    def get_achieved_goal(self) -> np.ndarray:
        """Return the achieved goal."""

    def get_goal(self) -> np.ndarray:
        """Return the current goal."""
        if self.goal is None:
            raise RuntimeError("No goal yet, call reset() first")
        else:
            return self.goal.copy()
    
    @abstractmethod
    def is_success(self, achieved_goal: np.ndarray, desired_goal: np.ndarray, info: Dict[str, Any]={}) -> np.ndarray:
        """Returns whether the achieved goal match the desired goal."""
    
    @abstractmethod
    def compute_reward(self, achieved_goal: np.ndarray, desired_goal: np.ndarray, info: Dict[str, Any] = {}) -> np.ndarray:
        """Compute reward associated to the achieved and the desired goal."""


from typing import Union
import numpy as np
from collections import namedtuple
from pathlib import Path

import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))
    
from rewards.reward_base import RewardBase


class SparseReward(RewardBase):

    def __init__(self, 
        is_potential: bool = False, 
        log_history_reward: bool=False,
        reward_constant: float=0.
    ) -> None:
        super().__init__(is_potential=is_potential, log_history_reward=log_history_reward)

        self.is_potential = is_potential
        
        self.reward_constant = reward_constant

    def get_reward(self, state: Union[namedtuple, np.ndarray], **kwargs) -> float:
        return self.reward_constant
    
    def reset(self):
        super().reset()



from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase


class ExtremeStateTermination(TerminationBase):

    def __init__(self, 
        v_max: float=1000., 
        p_max: float=500., 
        termination_reward: float=-1.,
        is_termination_reward_based_on_steps_left: bool=False,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        """飞机速度大于v_max，或者角速度不在[-p_max, p_max]范围内，判定飞机处于极限状态.

        Args:
            v_max (float, optional): 速度最大临界值。 Defaults to 1000..
            p_max (float, optional): 角速度最大临界值。 Defaults to 500..
        """
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=is_termination_reward_based_on_steps_left,
            env_config=env_config,
            my_logger=my_logger
        )
        self.v_max = v_max
        self.p_max = p_max
    
    def _get_termination(self, v: float, p: float) -> Tuple[bool, bool]:
        if v > self.v_max:
            return True, True
        if not (-self.p_max <= p <= self.p_max):
            return True, True
        return False, False

    def get_termination(self, state: namedtuple, **kwargs) -> Tuple[bool, bool]:
        assert 'v' in state._fields, "state中必须包含v"
        assert 'p' in state._fields, "state中必须包含p"
        v = state.v
        p = state.p
        return self._get_termination(v=v, p=p)

    def get_termination_and_reward(self, state: namedtuple, **kwargs) -> Tuple[bool, bool, float]:
        assert 'v' in state._fields, "state中必须包含v"
        assert 'p' in state._fields, "state中必须包含p"
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"

        v = state.v
        p = state.p
        terminated, truncated = self._get_termination(v=v, p=p)
        # reward = self.termination_reward if terminated else 0.

        return terminated, truncated, self.get_termination_penalty(terminated=terminated, steps_cnt=kwargs["step_cnt"])

    def reset(self):
        pass

    def __str__(self) -> str:
        return "extreme_state_termination"


from math import fabs, cos, sin, radians, sqrt, acos, degrees
import numpy as np
from enum import Enum


ROLL_THRESHOLD = 1e-2

class RollDirection(Enum):
    LEFT = 1
    RIGHT = 2
    NOROLL = 3


def roll2xy_on_unit_circle(roll):
    """将滚转角转换成单位圆上的坐标

    Args:
        roll (_type_): 单位度，[-180, 180]
    """

    return cos(radians(roll)), -sin(radians(roll))


def get_roll_direction(roll_1: float, roll_2: float, roll_threshold: float=ROLL_THRESHOLD):
    """判断roll_1到roll_2，飞机的滚转方向（按小于180度的角算转向！！！）.

    思路：

    平面直角坐标系：x正向表示滚转角0，y轴正向表示滚转角-90，y轴负向表示滚转角90，x轴负向上方-180，下方180

    将滚转角表示成半径为1的圆上的点，roll_1 -> (x1, y1)，roll_2 -> (x2, y2)

    通过计算(x1, y1)与(x2, y2)的叉积判断旋转方向（右手定则）

    注意：

    该方法计算的值按小于180度的角旋转计算的，例如：170右旋转到-5，旋转了185度。此方法是按170左旋175到-5，所以会返回左滚转！！！

    Args:
        roll_1 (float): deg, [-180, 180]
        roll_2 (float): deg, [-180, 180]
    """
    
    if fabs(roll_1 - roll_2) < roll_threshold:
        return RollDirection.NOROLL
    
    x1, y1 = roll2xy_on_unit_circle(roll_1)
    x2, y2 = roll2xy_on_unit_circle(roll_2)

    cross_product = x1 * y2 - y1 * x2

    if cross_product > 0.:
        return RollDirection.LEFT
    else:
        return RollDirection.RIGHT 


def get_roll_deg(roll_1: float, roll_2: float):
    """计算roll_1到roll_2滚转的角度（按小于180度的角算转向！！！）

    Args:
        roll_1 (float): _description_
        roll_2 (float): _description_

    Returns:
        _type_: _description_
    """
    x1, y1 = roll2xy_on_unit_circle(roll_1)
    x2, y2 = roll2xy_on_unit_circle(roll_2)

    inner_product = x1 * x2 + y1 * y2

    tmp = inner_product / (sqrt(x1 * x1 + y1 * y1) * sqrt(x2 * x2 + y2 * y2))
    tmp = np.clip(tmp, a_min=-1., a_max=1.)
    
    res_rad = acos(tmp)

    return degrees(res_rad)


#!/usr/local/bin/python
# -- coding: utf-8 --
''' config management tools
Created by Weijia on 2017/06/01
Last modified by Weijia on 2019/10/12
'''
import sys
if sys.version_info.major == 2:   # modified wh on 20200221
    import ConfigParser as configparser
else:
    import configparser
import string

def loadSecItems(filename, sec):
    cf = configparser.ConfigParser()
    cf.read(filename)
    items = cf.items(sec)    
    return items

def loadSecMap(filename, sec):
    cf = configparser.ConfigParser()
    cf.read(filename)
    items = cf.items(sec)
    map = {}
    for it in items:
        map[it[0]] = it[1]
    return map
    
def loadConfig(filename):
    cf = configparser.ConfigParser()
    cf.read(filename)
    section = cf.sections()
    map = {}
    for s in section:
        map[s] = {}
        items = cf.items(s)
        for ps in items:  
            map[s][ps[0]]=ps[1]
    return map

def judgeTrue(word):
    if word == None: return False
    return word.lower() == "true"


import numpy as np
from typing import Tuple

def angle_of_2_3d_vectors(v1, v2):
    x = np.array(v1)
    y = np.array(v2)

    # 分别计算两个向量的模：
    module_x = np.sqrt(x.dot(x))
    module_y = np.sqrt(y.dot(y))

    assert not np.any(np.allclose([module_x], [0.], atol=1e-8)), "v1的模为0！"
    assert not np.any(np.allclose([module_y], [0.], atol=1e-8)), "v2的模为0！"

    # 计算两个向量的点积
    dot_value = x.dot(y)

    # 计算夹角的cos值：
    cos_theta = dot_value / (module_x * module_y)
    # cos_theta在[-1, 1]之间
    cos_theta = np.clip(cos_theta, a_min=-1.0, a_max=1.0)

    # 求得夹角（弧度制）：
    angle_radian = np.arccos(cos_theta)

    # 转换为角度值：
    # angle_value = angle_radian * 180. / np.pi 
    angle_value = np.rad2deg(angle_radian)
    return angle_value

def angle_of_2_velocity(v_1: float, mu_1: float, chi_1: float, v_2: float, mu_2: float, chi_2: float):
    """计算两个速度矢量之间的夹角，两个速度矢量都由(v, mu, chi)描述。
    """
    current_velocity_vector = [
        v_1 * np.cos(np.deg2rad(mu_1)) * np.sin(np.deg2rad(chi_1)), 
        v_1 * np.cos(np.deg2rad(mu_1)) * np.cos(np.deg2rad(chi_1)),
        v_1 * np.sin(np.deg2rad(mu_1)),
    ]
    target_velocity_vector = [
        v_2 * np.cos(np.deg2rad(mu_2)) * np.sin(np.deg2rad(chi_2)), 
        v_2 * np.cos(np.deg2rad(mu_2)) * np.cos(np.deg2rad(chi_2)),
        v_2 * np.sin(np.deg2rad(mu_2)),
    ]
    return angle_of_2_3d_vectors(current_velocity_vector, target_velocity_vector)

def angle_of_2_velocity2(ve_1: float, vn_1: float, vh_1: float, v_2: float, mu_2: float, chi_2: float):
    """计算两个速度矢量之间的夹角，第一个速度矢量由(ve, vn, vh)描述，第二个速度矢量由(v, mu, chi)描述
    """
    current_velocity_vector = [
        ve_1,
        vn_1,
        vh_1
    ]
    target_velocity_vector = [
        v_2 * np.cos(np.deg2rad(mu_2)) * np.sin(np.deg2rad(chi_2)), 
        v_2 * np.cos(np.deg2rad(mu_2)) * np.cos(np.deg2rad(chi_2)),
        v_2 * np.sin(np.deg2rad(mu_2)),
    ]
    return angle_of_2_3d_vectors(current_velocity_vector, target_velocity_vector)

def v_mu_chi_2_enh(v: float, mu: float, chi: float) -> Tuple[float, float, float]:
    """航迹角表示的速度矢量转“东北天（east-north-height，右手坐标系）”坐标

    Args:
        v (float): 速度
        mu (float): 航迹倾斜角，单位：度
        chi (float): 航迹方位角，单位：度

    Returns:
        (float, float, float): (east, north, height)
    """
    mu_rad = np.deg2rad(mu)
    chi_rad = np.deg2rad(chi)
    return (
        v * np.cos(mu_rad) * np.sin(chi_rad),
        v * np.cos(mu_rad) * np.cos(chi_rad),
        v * np.sin(mu_rad)
    )


from abc import ABC, abstractmethod
from typing import Tuple
from collections import namedtuple
import logging
import numpy as np


class TerminationBase(ABC):

    def __init__(self, 
        termination_reward: float=-1.,
        is_termination_reward_based_on_steps_left: bool=False,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        """
        Args:
            termination_reward (float, optional): 结束奖励. Defaults to -1..
            is_termination_reward_based_on_steps_left (bool, optional): 计算结束奖励时，是否基于剩余的仿真步数。不是的话，结束奖励就是termination_reward；是的话，结束奖励按剩余每一步的奖励是-1计算，因此累计起来是(-1) * (1 - gamma^{step_left})/(1-gamma). Defaults to False.
            max_episode_steps (int, optional): 环境的最大仿真步长. Defaults to 400.
            my_logger (logging.Logger, optional): _description_. Defaults to None.
        """
        self.termination_reward = termination_reward
        self.is_termination_reward_based_on_steps_left = is_termination_reward_based_on_steps_left
        self.env_config = env_config
        self.logger = my_logger

    @abstractmethod
    def get_termination(self, state: namedtuple, **kwargs) -> Tuple[bool, bool]:
        """返回episode是否结束，是否被截断

        Returns:
            Tuple[bool, bool]: terminated, truncated
        """
        raise NotImplementedError
    
    @abstractmethod
    def get_termination_and_reward(self, state: namedtuple, **kwargs) -> Tuple[bool, bool, float]:
        """返回episode是否结束，是否被截断，以及结束的对应的奖励

        Args:
            state (_type_): _description_

        Returns:
            Tuple[float, bool, bool]: _description_
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self):
        raise NotImplementedError

    def get_penalty_base_on_steps_left(self, steps_cnt: int=1):
        """用于环境期望仿真长度固定的情况。
        
        当触发了提前终止条件后，假设后续每一步的奖励是-1，

        累加起来就是 - (1 - gamma^steps_left) / (1 - gamma).

        Args:
            steps_cnt (int, optional): 环境已经仿真的步数. Defaults to 1.

        Returns:
            _type_: _description_
        """
        return - (1 -  np.power(self.rl_gamma, self.max_episode_steps - steps_cnt)) / (1 - self.rl_gamma)
    
    def get_termination_penalty(self, terminated: bool=False, steps_cnt: int=1):
        """计算仿真结束对应的惩罚值。

        Args:
            terminated (bool, optional): 仿真是否结束. Defaults to False.
            steps_cnt (int, optional): 环境已经仿真的步数. Defaults to 1.

        Returns:
            _type_: _description_
        """
        if terminated:
            if self.is_termination_reward_based_on_steps_left:
                reward = self.get_penalty_base_on_steps_left(steps_cnt)
            else:
                reward = self.termination_reward
        else:
            reward = 0.
        
        return reward
    
    @property
    def rl_gamma(self):
        return self.env_config["task"].get("gamma")

    @property
    def step_frequence(self):
        return self.env_config["task"].get("step_frequence")

    @property
    def max_simulate_time(self):
        return self.env_config["task"].get("max_simulate_time")

    @property
    def max_episode_steps(self):
        return self.max_simulate_time * self.step_frequence


from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase


class ReachTargetTerminationSingleStep(TerminationBase):

    def __init__(self,
        v_threshold: float=10., 
        phi_threshold: float=1.,
        theta_threshold: float=1.,
        termination_reward: float=1.,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=False,
            env_config=env_config, 
            my_logger=my_logger
        )
        
        self.v_threshold = v_threshold
        self.phi_threshold = phi_threshold
        self.theta_threshold = theta_threshold

    def _get_termination(
        self, 
        v: float, phi: float, theta: float, 
        goal_v: float, goal_phi: float, goal_theta: float
    ) -> Tuple[bool, bool]:
        
        v_flag, phi_flag, theta_flag = False, False, False
        if abs(goal_v - v) < self.v_threshold:
            v_flag = True
        if abs(goal_phi - phi) < self.phi_threshold:
            phi_flag = True
        if abs(goal_theta - theta) < self.theta_threshold:
            theta_flag = True
        
        if v_flag and phi_flag and theta_flag:
            return True, False
        else:
            return False, False

    def get_termination(self, state, **kwargs) -> Tuple[bool, bool]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_phi" in kwargs, "参数中需要包括goal_phi，再调用get_termination方法"
        assert "goal_theta" in kwargs, "参数中需要包括goal_theta，再调用get_termination方法"

        return self._get_termination(
            v=state.v,
            phi=state.phi,
            theta=state.theta,
            goal_v=kwargs["goal_v"], 
            goal_phi=kwargs["goal_phi"], 
            goal_theta=kwargs["goal_theta"],
        )

    def get_termination_and_reward(self, state, **kwargs) -> Tuple[bool, bool, float]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_phi" in kwargs, "参数中需要包括goal_phi，再调用get_termination方法"
        assert "goal_theta" in kwargs, "参数中需要包括goal_theta，再调用get_termination方法"

        terminated, truncated = self._get_termination(
            v=state.v,
            phi=state.phi,
            theta=state.theta,
            goal_v=kwargs["goal_v"], 
            goal_phi=kwargs["goal_phi"], 
            goal_theta=kwargs["goal_theta"],
        )
        reward = self.termination_reward if terminated else 0.
        
        return terminated, truncated, reward

    def reset(self):
        pass

    def __str__(self) -> str:
        return "reach_target_termination_single_step_based_on_v_phi_theta"


from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase


class NegativeOverloadAndBigPhiTermination(TerminationBase):
    """负过载且大幅度滚转：nz<0且phi>60(deg)连续超过2秒
    """
    def __init__(self,
        time_window: float=2.,
        negative_overload_threshold: float=0.,
        big_phi_threshold: float=60.,
        termination_reward: float=-1.,
        is_termination_reward_based_on_steps_left: bool=False,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=is_termination_reward_based_on_steps_left,
            env_config=env_config,
            my_logger=my_logger
        )

        self.time_window = time_window
        self.negative_overload_threshold = negative_overload_threshold
        self.big_phi_threshold = big_phi_threshold

        self.invalid_cnt = 0

    def _get_termination(self, state: namedtuple, nz: float) -> Tuple[bool, bool]:
        if abs(state.phi) > self.big_phi_threshold and nz < self.negative_overload_threshold:
            self.invalid_cnt += 1
        else:
            self.invalid_cnt = 0
        
        if self.invalid_cnt >= self.time_window_step_length:
            return True, False
        
        return False, False
    
    def get_termination(self, state: namedtuple, **kwargs) -> Tuple[bool, bool]:
        assert "nz" in kwargs, "参数中需要包括nz，再调用get_termination方法"

        return self._get_termination(state=state, nz=kwargs["nz"])

    def get_termination_and_reward(self, state: namedtuple, **kwargs) -> Tuple[bool, bool, float]:
        assert "nz" in kwargs, "参数中需要包括nz，再调用get_termination方法"
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"

        terminated, truncated = self._get_termination(state=state, nz=kwargs["nz"])
        # reward = self.termination_reward if terminated else 0.
        
        return terminated, truncated, self.get_termination_penalty(terminated=terminated, steps_cnt=kwargs["step_cnt"])

    
    def reset(self):
        self.invalid_cnt = 0

    @property
    def time_window_step_length(self) -> int:
        """判断使用的时间窗口step数

        Returns:
            _type_: _description_
        """
        return round(self.time_window * self.step_frequence)

    def __str__(self) -> str:
        return "negative_overload_and_big_phi_termination"


import logging
from pathlib import Path

PROJECT_ROOT_DIR = Path(__file__).parent.parent

def get_logger(logger_name:str="ucav", log_file_dir: Path=PROJECT_ROOT_DIR / "my_logs" / "my_sys_logs.log"):
    handler_file = logging.FileHandler(log_file_dir)  # stdout to file
    handler_control = logging.StreamHandler()    # stdout to console
    handler_file.setLevel('DEBUG')               # 设置级别: DEBUG 、INFO 、WARNING、ERROR、CRITICAL
    handler_control.setLevel('DEBUG')             

    selfdef_fmt = '%(asctime)s - %(funcName)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(selfdef_fmt)
    handler_file.setFormatter(formatter)
    handler_control.setFormatter(formatter)

    logger = logging.getLogger(logger_name)
    logger.setLevel('DEBUG')           #设置了这个才会把debug以上的输出到控制台

    logger.addHandler(handler_file)    #添加handler
    logger.addHandler(handler_control)

    return logger

# my_logger = get_logger()

# def get_default_logger() -> logging.Logger:
#     return my_logger


if __name__ == "__main__":
    my_logger = get_logger()
    my_logger.info('info,一般的信息输出')
    my_logger.warning('waring，用来用来打印警告信息')
    my_logger.error('error，一般用来打印一些错误信息')
    my_logger.critical('critical，用来打印一些致命的错误信息，等级最高')


from typing import Union
import numpy as np
from collections import namedtuple
from pathlib import Path

import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))
    
from rewards.reward_base import RewardBase


class SparseReward2(RewardBase):

    def __init__(self, 
        is_potential: bool = False, log_history_reward: bool=False,
        step_frequence: int=100, integral_time_length: float=1.,
        v_threshold: float=10., mu_threshold: float=1., chi_threshold: float=1.,
        reach_target_reward: float=0., else_reward: float=0. 
    ) -> None:
        super().__init__(is_potential=is_potential, log_history_reward=log_history_reward)

        self.is_potential = is_potential
        
        self.step_frequence = step_frequence
        self.integral_time_length = integral_time_length
        
        self.v_threshold = v_threshold
        self.mu_threshold = mu_threshold
        self.chi_threshold = chi_threshold

        self.reach_target_reward = reach_target_reward
        self.else_reward = else_reward

    def get_reward(self, state: Union[namedtuple, np.ndarray], **kwargs) -> float:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_mu" in kwargs, "参数中需要包括goal_mu，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        assert "state_list" in kwargs, "参数中需要包括state_list（list[namedtuple]类型，所有历史观测），再调用get_reward方法"
        assert type(kwargs["state_list"]) is list, "state_list参数的类型应该是list[namedtuple]"

        state_list = kwargs["state_list"]  # list[namedtuple], GuideEnv.get_state_vars()返回的namedtuple

        if len(state_list) < self.integral_window_length:
            return self.else_reward
        else:
            v_flag, mu_flag, chi_flag = False, False, False
            if sum([abs(kwargs["goal_v"] - item.v) for item in state_list[-self.integral_window_length:]]) < self.v_integral_threshold:
                v_flag = True
            if sum([abs(kwargs["goal_mu"] - item.mu) for item in state_list[-self.integral_window_length:]]) < self.mu_integral_threshold:
                mu_flag = True
            if sum([abs(kwargs["goal_chi"] - item.chi) for item in state_list[-self.integral_window_length:]]) < self.chi_integral_threshold:
                chi_flag = True
            if v_flag and mu_flag and chi_flag:
                return self.reach_target_reward
            else:
                return self.else_reward
    
    def reset(self):
        super().reset()

    @property
    def integral_window_length(self):
        """v, mu, chi的积分窗口长度

        Returns:
            _type_: _description_
        """
        return self.integral_time_length * self.step_frequence

    @property
    def v_integral_threshold(self):
        """v的误差积分阈值，当v在最后self.integral_window_length上的积分小于该值时，认为v达到目标值

        Returns:
            _type_: _description_
        """
        return self.v_threshold * self.integral_window_length
    
    @property
    def mu_integral_threshold(self):
        """mu的误差积分阈值，当mu在最后self.integral_window_length上的积分小于该值时，认为mu达到目标值

        Returns:
            _type_: _description_
        """
        return self.mu_threshold * self.integral_window_length
    
    @property
    def chi_integral_threshold(self):
        """chi的误差积分阈值，当chi在最后self.integral_window_length上的积分小于该值时，认为chi达到目标值

        Returns:
            _type_: _description_
        """
        return self.chi_threshold * self.integral_window_length


from typing import Tuple
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase

class TimeoutTermination(TerminationBase):
    """超时：40秒内未达到目标速度矢量
    """
    def __init__(self, 
        termination_reward: float=-1.,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=False, 
            env_config=env_config, 
            my_logger=my_logger
        )

    def _get_termination(self, step_cnt: int) -> Tuple[bool, bool]:
        if step_cnt >= self.max_episode_steps - 1:
            return True, True

        return False, False

    def get_termination(self, state, **kwargs) -> Tuple[bool, bool]:
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"
        assert type(kwargs["step_cnt"]) is int, "step_cnt必须为int类型"

        return self._get_termination(step_cnt=kwargs["step_cnt"])
    
    def get_termination_and_reward(self, state, **kwargs) -> Tuple[bool, bool, float]:
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"
        assert type(kwargs["step_cnt"]) is int, "step_cnt必须为int类型"

        terminated, truncated = self._get_termination(step_cnt=kwargs["step_cnt"])
        # reward = self.termination_reward if terminated else 0.
        
        return terminated, truncated, self.get_termination_penalty(terminated=terminated, steps_cnt=kwargs["step_cnt"])
        
    def reset(self):
        pass

    def __str__(self) -> str:
        return "timeout_termination"


from abc import ABC, abstractmethod
import numpy as np
from collections import namedtuple
from typing import Union, Callable
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from rewards.reward_base import RewardBase


class PonentialRewardBase(RewardBase, ABC):

    def __init__(self, gamma: float=0.99, my_logger: logging.Logger=None) -> None:
        super().__init__(is_potential=False, log_history_reward=True, my_logger=my_logger)  # 直接计算基于势的奖励，不使用基类的默认逻辑
        self.gamma = gamma
    
    def get_reward(self, state: namedtuple, **kwargs) -> float:
        """根据（state, next_state, done）计算ponential reward。
        
        参考论文：“Reward Shaping in Episodic Reinforcement Learning”
        
        计算公式为：\gamma * \Phi(next_state) - \Phi(state), 
        
        其中需要特殊处理的是，当done=True时，\Phi(next_state)取值为0！！！
        """
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_mu" in kwargs, "参数中需要包括goal_mu，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        assert "next_state" in kwargs, "调用前，需要把namedtuple类型的next_state放入additional_info"
        assert "done" in kwargs, ""

        next_state = kwargs["next_state"]
        done = kwargs["done"]

        reward_v = self.gamma * (0. if done else self.phi_v(next_state, kwargs["goal_v"])) - self.phi_v(state, kwargs["goal_v"])
        reward_mu = self.gamma * (0. if done else self.phi_mu(next_state, kwargs["goal_mu"])) - self.phi_mu(state, kwargs["goal_mu"])
        reward_chi = self.gamma * (0. if done else self.phi_chi(next_state, kwargs["goal_chi"])) - self.phi_chi(state, kwargs["goal_chi"])
        reward = reward_v * reward_mu * reward_chi

        reward_log = {
            "reward_v": reward_v,
            "reward_mu": reward_mu,
            "reward_chi": reward_chi,
            "reward": reward
        }

        return self._process(new_reward=reward, log=reward_log)

    def reset(self):
        super().reset()

    @abstractmethod
    def phi_v(self, state: namedtuple, goal_v: float):
        raise NotImplementedError

    @abstractmethod
    def phi_mu(self, state: namedtuple, goal_mu: float):
        raise NotImplementedError
    
    @abstractmethod
    def phi_chi(self, state: namedtuple, goal_chi: float):
        raise NotImplementedError


class PonentialReward1(PonentialRewardBase):
    """计算势的方法：

    - abs(target - state) / (state.max - state.min)

    即：目标状态处势为0,其他状态的势为负数，值根据误差和归一化因子计算。

    Args:
        PonentialRewardBase (_type_): _description_
    """

    def __init__(self, gamma: float = 0.99, 
        v_min: float=0., v_max: float=1000., 
        mu_min: float=-90., mu_max: float=90., 
        chi_min: float=-180., chi_max: float=180.,
    ) -> None:
        super().__init__(gamma=gamma)

        self.v_min, self.v_max = v_min, v_max
        self.mu_min, self.mu_max = mu_min, mu_max
        self.chi_min, self.chi_max = chi_min, chi_max        

    def phi_v(self, state: namedtuple, goal_v: float):
        return -abs(goal_v - state.v) / (self.v_max - self.v_min)

    def phi_mu(self, state: namedtuple, goal_mu: float):
        return -abs(goal_mu - state.mu) / (self.mu_max - self.mu_min)

    def phi_chi(self, state: namedtuple, goal_chi: float):
        return -abs(goal_chi - state.chi) / (self.chi_max - self.chi_min)
    
    def __str__(self) -> str:
        return "ponential_reward_1"


class PonentialReward2(PonentialRewardBase):
    """计算势的方法，参考了论文“Autonomous Control of Simulated Fixed Wing Aircraft using Deep Reinforcement Learning”的3.6.3节
    """

    def __init__(self, 
        gamma: float = 0.99,
        coef_k_for_v: float=10., coef_k_for_mu: float=5., coef_k_for_chi: float=5.
    ) -> None:
        """_summary_

        Args:
            gamma (float, optional): 同MDP中的gamma. Defaults to 0.99.
            coef_k_for_v (float, optional): 系数越小，对误差越敏感. Defaults to 10..
            coef_k_for_mu (float, optional): 系数越小，对误差越敏感. Defaults to 5..
            coef_k_for_chi (float, optional): 系数越小，对误差越敏感. Defaults to 5..
        """
        super().__init__(gamma=gamma)

        self.coef_k_for_v, self.coef_k_for_mu, self.coef_k_for_chi = coef_k_for_v, coef_k_for_mu, coef_k_for_chi

    def phi_v(self, state: namedtuple, goal_v: float):
        err = abs(goal_v - state.v) / self.coef_k_for_v  # 计算误差
        normalized_err =  err / (1 + err)  # 将误差缩放到[0, 1]
        return 1. - normalized_err

    def phi_mu(self, state: namedtuple, goal_mu: float):
        err = abs(goal_mu - state.mu) / self.coef_k_for_mu  # 计算误差
        normalized_err =  err / (1 + err)  # 将误差缩放到[0, 1]
        return 1. - normalized_err
    
    def phi_chi(self, state: namedtuple, goal_chi: float):
        err = abs(goal_chi - state.chi) / self.coef_k_for_chi  # 计算误差
        normalized_err =  err / (1 + err)  # 将误差缩放到[0, 1]
        return 1. - normalized_err

    def __str__(self) -> str:
        return "ponential_reward_2"
    

class ScaledPonentialReward2(PonentialReward2):
    """将PonentialReward2的势函数计算的值放大scale_coef倍，尝试解决PonentialReward2计算的奖励数值过小的问题

    Args:
        PonentialReward2 (_type_): _description_
    """
    def __init__(self, scale_coef: float=100., gamma: float = 0.99, coef_k_for_v: float = 10, coef_k_for_mu: float = 5, coef_k_for_chi: float = 5) -> None:
        super().__init__(gamma, coef_k_for_v, coef_k_for_mu, coef_k_for_chi)
        self.scale_coef = scale_coef
    
    def phi_v(self, state: namedtuple, goal_v: float):
        return self.scale_coef * super().phi_v(state, goal_v)

    def phi_mu(self, state: namedtuple, goal_mu: float):
        return self.scale_coef * super().phi_mu(state, goal_mu)
    
    def phi_chi(self, state: namedtuple, goal_chi: float):
        return self.scale_coef * super().phi_chi(state, goal_chi)
    
    def __str__(self) -> str:
        return "scaled_ponential_reward_2"
    



def update_nested_dict(d1, d2):
    for key, value in d2.items():
        if key in d1 and isinstance(d1[key], dict) and isinstance(value, dict):
            update_nested_dict(d1[key], value)
        else:
            d1[key] = value


from abc import ABC, abstractmethod
import numpy as np
from collections import namedtuple
from typing import Union, Callable
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from rewards.reward_base import RewardBase
from utils.geometry_utils import angle_of_2_3d_vectors


class PonentialRewardBasedOnAngle(RewardBase, ABC):
    """基于速度矢量方向的shaping reward

    Args:
        RewardBase (_type_): _description_
        ABC (_type_): _description_
    """

    def __init__(self, b: float=1., gamma: float=0.99, log_history_reward: bool = True, my_logger: logging.Logger = None) -> None:
        self.b = b
        self.gamma = gamma
        super().__init__(is_potential=False, log_history_reward=log_history_reward, my_logger=my_logger)
    
    def get_reward(self, state: Union[namedtuple, np.ndarray], **kwargs) -> float:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_mu" in kwargs, "参数中需要包括goal_mu，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        # assert "ve" in kwargs, "调用前，需要传入ve"
        # assert "vn" in kwargs, "调用前，需要传入vn"
        # assert "vh" in kwargs, "调用前，需要传入vh"
        # assert "next_ve" in kwargs, "调用前，需要传入next_ve"
        # assert "next_vn" in kwargs, "调用前，需要传入next_vn"
        # assert "next_vh" in kwargs, "调用前，需要传入next_vh"
        assert "next_state" in kwargs, "调用前，需要把namedtuple类型的next_state放入函数的参数中"
        assert "done" in kwargs, ""

        next_state = kwargs["next_state"]
        done = kwargs["done"]
        # ve, vn, vh = kwargs["ve"], kwargs["vn"], kwargs["vh"]
        # next_ve, next_vn, next_vh = kwargs["next_ve"], kwargs["next_vn"], kwargs["next_vh"]

        reward = self.gamma * (0. if done else self.phi(next_state, kwargs["goal_v"], kwargs["goal_mu"], kwargs["goal_chi"])) - self.phi(state, kwargs["goal_v"], kwargs["goal_mu"], kwargs["goal_chi"])
        
        return self._process(new_reward=reward)

    def phi(self, state: namedtuple, goal_v: float, goal_mu: float, goal_chi: float):
        plane_current_velocity_vector = [
            state.v * np.cos(np.deg2rad(state.mu)) * np.sin(np.deg2rad(state.chi)), 
            state.v * np.cos(np.deg2rad(state.mu)) * np.cos(np.deg2rad(state.chi)),
            state.v * np.sin(np.deg2rad(state.mu)),
        ]

        target_velocity_vector = [
            goal_v * np.cos(np.deg2rad(goal_mu)) * np.sin(np.deg2rad(goal_chi)), 
            goal_v * np.cos(np.deg2rad(goal_mu)) * np.cos(np.deg2rad(goal_chi)),
            goal_v * np.sin(np.deg2rad(goal_mu)),
        ]

        angle = angle_of_2_3d_vectors(plane_current_velocity_vector, target_velocity_vector)

        return -np.power(angle / 180., self.b)

    def reset(self):
        super().reset()


from typing import Tuple, List
from math import fabs
from collections import namedtuple
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from utils.attitude_angle_calc_utils import RollDirection, get_roll_direction, get_roll_deg
from terminations.termination_base import TerminationBase

class ContinuouselyRollTermination(TerminationBase):
    """连续滚转：连续滚转2圈
    """
    def __init__(self,
        continuousely_roll_threshold: float=720.,
        termination_reward: float=-1.,
        is_termination_reward_based_on_steps_left: bool=False,
        env_config: dict=None,
        my_logger: logging.Logger = None) -> None:
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=is_termination_reward_based_on_steps_left,
            env_config=env_config,
            my_logger=my_logger
        )

        self.continuousely_roll_threshold = continuousely_roll_threshold

        self.roll_flag: RollDirection = RollDirection.NOROLL
        self.accumulate_roll = 0.

    def _get_termination(self, state: namedtuple, next_state: namedtuple) -> Tuple[bool, bool]:
        roll_1, roll_2 = state.phi, next_state.phi
        cur_roll_direction = get_roll_direction(roll_1, roll_2)
        cur_roll_deg = get_roll_deg(roll_1=roll_1, roll_2=roll_2)
        # print(f"roll: {roll_1} -> {roll_2}, direction: {cur_roll_direction}, pre_roll: {self.roll_flag}, sum: {self.accumulate_roll}")
        if cur_roll_direction == RollDirection.NOROLL:
            self.reset()
            return False, False
        else:
            if cur_roll_direction == self.roll_flag:
                # TODO: 需要处理的边界情况：(-180) - 170，通过叉积求旋转角度
                self.accumulate_roll += cur_roll_deg
                if self.accumulate_roll > self.continuousely_roll_threshold:
                    if self.logger is not None:
                        self.logger.info(f"持续转弯超过{self.continuousely_roll_threshold}度！！！")
                    return True, False
                else:
                    return False, False
            else:
                self.accumulate_roll = cur_roll_deg
                self.roll_flag = cur_roll_direction
                return False, False

    def get_termination(self, state: namedtuple, **kwargs) -> Tuple[bool, bool]:
        assert "next_state" in kwargs, "参数中需要包括next_state，再调用get_termination方法"

        return self._get_termination(state=state, next_state=kwargs["next_state"])
    
    def get_termination_and_reward(self, state: namedtuple, **kwargs) -> Tuple[bool, bool, float]:
        assert "next_state" in kwargs, "参数中需要包括next_state，再调用get_termination方法"
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"

        terminated, truncated = self._get_termination(state=state, next_state=kwargs["next_state"])
        # reward = self.termination_reward if terminated else 0.

        return terminated, truncated, self.get_termination_penalty(terminated=terminated, steps_cnt=kwargs["step_cnt"])

    def reset(self):
        self.roll_flag = RollDirection.NOROLL
        self.accumulate_roll = 0.
    
    def __str__(self) -> str:
        return "continuousely_roll_termination"



from typing import Union
import numpy as np
from collections import namedtuple
from pathlib import Path
import logging
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from rewards.reward_base import RewardBase
from utils.geometry_utils import angle_of_2_3d_vectors


class DenseRewardBasedOnAngleAndVelocity(RewardBase):
    """基于飞机速度矢量的大小与方向的过程奖励。

    Args:
        RewardBase (_type_): _description_
    """
    def __init__(
            self, 
            b: float = 1., 
            angle_scale: float = 180., 
            velocity_scale: float = 100.,
            angle_weight: float = 0.5, 
            log_history_reward: bool = True, 
            my_logger: logging.Logger = None
        ) -> None:
        assert 0. <= angle_weight <= 1., "angle_weight不在[0,1]范围内"

        self.b = b
        self.angle_scale = angle_scale
        self.velocity_scale = velocity_scale
        self.angle_weight = angle_weight
        super().__init__(is_potential=False, log_history_reward=log_history_reward, my_logger=my_logger)
    
    def get_reward(self, state: Union[namedtuple, np.ndarray], **kwargs) -> float:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_mu" in kwargs, "参数中需要包括goal_mu，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        # assert "next_ve" in kwargs, "调用前，需要传入next_ve"
        # assert "next_vn" in kwargs, "调用前，需要传入next_vn"
        # assert "next_vh" in kwargs, "调用前，需要传入next_vh"
        assert "next_state" in kwargs, "调用前，需要把namedtuple类型的next_state放入函数的参数中"

        next_state = kwargs["next_state"]

        # plane_current_velocity_vector = [
        #     kwargs["next_ve"], 
        #     kwargs["next_vn"], 
        #     kwargs["next_vh"]
        # ]
        plane_current_velocity_vector = [
            next_state.v * np.cos(np.deg2rad(next_state.mu)) * np.sin(np.deg2rad(next_state.chi)), 
            next_state.v * np.cos(np.deg2rad(next_state.mu)) * np.cos(np.deg2rad(next_state.chi)), 
            next_state.v * np.sin(np.deg2rad(next_state.mu))
        ]
        target_velocity_vector = [
            kwargs["goal_v"] * np.cos(np.deg2rad(kwargs["goal_mu"])) * np.sin(np.deg2rad(kwargs["goal_chi"])), 
            kwargs["goal_v"] * np.cos(np.deg2rad(kwargs["goal_mu"])) * np.cos(np.deg2rad(kwargs["goal_chi"])),
            kwargs["goal_v"] * np.sin(np.deg2rad(kwargs["goal_mu"])),
        ]
        angle = angle_of_2_3d_vectors(plane_current_velocity_vector, target_velocity_vector)
        angle_base_reward = - np.power(angle / self.angle_scale, self.b)

        # 奖励公式参考论文“Reinforcement learning for UAV attitude control”的4.3节
        velocity_error = np.abs(kwargs["goal_v"] - next_state.v)
        cliped_velocity_error = np.clip(velocity_error / self.velocity_scale, a_min=0., a_max=1.)
        velocity_base_reward = - np.power(cliped_velocity_error, self.b)

        return self.angle_weight *  angle_base_reward + (1. - self.angle_weight) * velocity_base_reward

    def reset(self):
        super().reset()


from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase


class CrashTermination(TerminationBase):

    def __init__(self, 
        h0: float=0., 
        termination_reward: float=-1.,
        is_termination_reward_based_on_steps_left: bool=False,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        """

        Args:
            h0 (float, optional): 飞机高度小于h0，判定坠机. Defaults to 0..
        """
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=is_termination_reward_based_on_steps_left,
            env_config=env_config,
            my_logger=my_logger
        )
        self.h0 = h0
    
    def _get_termination(self, h: float) -> Tuple[bool, bool]:
        if h < self.h0:
            return True, True
        return False, False

    def get_termination(self, state: namedtuple, **kwargs) -> Tuple[bool, bool]:
        assert 'h' in state._fields, "state中必须包含h"
        h = state.h
        return self._get_termination(h=h)

    def get_termination_and_reward(self, state: namedtuple, **kwargs) -> Tuple[bool, bool, float]:
        assert 'h' in state._fields, "state中必须包含h"
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"

        h = state.h
        terminated, truncated = self._get_termination(h=h)
        # reward = self.termination_reward if terminated else 0.
        return terminated, truncated, self.get_termination_penalty(terminated=terminated, steps_cnt=kwargs["step_cnt"])

    def reset(self):
        pass

    def __str__(self) -> str:
        return "crash_termination"


import sys
from pathlib import Path
import numpy as np
import pandas as pd
from gymnasium.utils import seeding

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from utils.load_config import load_config
from utils.dict_utils import update_nested_dict

class GoalSampler(object):

    def __init__(
        self, 
        np_random: np.random.Generator, 
        env_config: dict
    ) -> None:
        self.env_config: dict = env_config

        self.flag_str = self.env_config.get("flag_str", "Train")
        self.debug_mode = self.env_config.get("debug_mode", True)

        self.np_random: np.random.Generator = np_random

        self.use_fixed_goal = self.env_config["goal"].get("use_fixed_goal")
        self.goal_v = self.env_config["goal"].get("goal_v")
        self.goal_mu = self.env_config["goal"].get("goal_mu")
        self.goal_chi = self.env_config["goal"].get("goal_chi")
        self.goal_expert_length = 0

        self.sample_random = self.env_config["goal"].get("sample_random")  # does sample randomly
        if not self.use_fixed_goal and not self.sample_random:
            self._load_available_goals()
            
            # does only sample reachable goals (length > 0) from available goals
            self.sample_reachable_goal = self.env_config["goal"].get("sample_reachable_goal")
            if self.sample_reachable_goal:
                self.available_goals = self.available_goals[self.available_goals.length > 0]
            
            self.sample_goal_noise_std = self.env_config["goal"].get("sample_goal_noise_std")

    def _load_available_goals(self):
        self.available_goals_file: str = self.env_config["goal"].get("available_goals_file")
        tmp_file_path = Path(self.available_goals_file)
        
        # self.available_goals are pd.DataFrame with columns: goal_v, goal_mu, goal_chi, length
        # length represents the trajectory length of reaching (goal_v, goal_mu, goal_chi)
        self.available_goals = pd.read_csv(tmp_file_path.absolute())
        self.available_goals = self.available_goals[["v", "mu", "chi", "length"]]

    def sample_goal(self) -> dict:
        if not self.use_fixed_goal:
            if self.sample_random:
                self.goal_v = self.np_random.uniform(low=self.env_config["goal"]["v_min"], high=self.env_config["goal"]["v_max"])
                self.goal_mu = self.np_random.uniform(low=self.env_config["goal"]["mu_min"], high=self.env_config["goal"]["mu_max"])
                self.goal_chi = self.np_random.uniform(low=self.env_config["goal"]["chi_min"], high=self.env_config["goal"]["chi_max"])

                # if self.debug_mode:
                #     print(f"In sampler, sample randomly: {self.goal_v}, {self.goal_mu}, {self.goal_chi}")
            else:
                if self.sample_reachable_goal:
                    while True:
                        tmp_goal = self.np_random.choice(self.available_goals, 1).squeeze()
                        if tmp_goal[3] == 0:
                            continue
                        else:
                            sampled_noise = self.sample_noise()

                            self.goal_v = tmp_goal[0] + sampled_noise[0]
                            self.goal_mu = tmp_goal[1] + sampled_noise[1]
                            self.goal_chi = tmp_goal[2] + sampled_noise[2]
                            self.goal_expert_length = tmp_goal[3]
                            break

                    # if self.debug_mode:
                    #     print(f"In sampler: sample reachable goal: {self.goal_v}, {self.goal_mu}, {self.goal_chi}, {self.goal_expert_length}")
                else:
                    tmp_goal = self.np_random.choice(self.available_goals, 1).squeeze()
                    sampled_noise = self.sample_noise()

                    self.goal_v = tmp_goal[0] + sampled_noise[0]
                    self.goal_mu = tmp_goal[1] + sampled_noise[1]
                    self.goal_chi = tmp_goal[2] + sampled_noise[2]
                    self.goal_expert_length = tmp_goal[3]

                    # if self.debug_mode:
                    #     print(f"In sampler, sample available goal: {self.goal_v}, {self.goal_mu}, {self.goal_chi}, {self.goal_expert_length}")

        return {
            "v": self.goal_v,
            "mu": self.goal_mu,
            "chi": self.goal_chi
        }
    
    def sample_noise(self):
        """采样噪声
        """
        return (
            (2 * self.np_random.random() -1.) * self.sample_goal_noise_std[0], 
            (2 * self.np_random.random() -1.) * self.sample_goal_noise_std[1], 
            (2 * self.np_random.random() -1.) * self.sample_goal_noise_std[2]
        )


from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Union
import numpy as np
from typing import Any
import logging


class RewardBase(ABC):

    def __init__(
        self, 
        is_potential: bool=False, 
        log_history_reward: bool=False, 
        my_logger: logging.Logger=None
    ) -> None:
        self.is_potential = is_potential  # 奖励是否基于势
        self.pre_reward = 0.  # 上一回合的奖励值，如果奖励基于势能，该值记录势
        self.reward_trajectory = []
        self.log_history_reward = log_history_reward
        self.logger = my_logger
    
    @abstractmethod
    def reset(self):
        self.pre_reward = 0.
        self.reward_trajectory = []

    @abstractmethod
    def get_reward(self, state: Union[namedtuple, np.ndarray], **kwargs) -> float:
        raise NotImplementedError

    def _process(self, new_reward, log: Any=None):
        """根据是否基于势能做额外处理，并记录
        """
        reward = new_reward
        
        if self.is_potential:
            reward, self.pre_reward = new_reward - self.pre_reward, new_reward
        
        if self.log_history_reward:
            if log is None:
                self.reward_trajectory.append(reward)
            else: 
                self.reward_trajectory.append(log)
        
        return reward





from typing import Union
import numpy as np
from collections import namedtuple
from pathlib import Path
import logging
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from rewards.reward_base import RewardBase
from utils.geometry_utils import angle_of_2_3d_vectors


class DenseReward(RewardBase):
    """过程奖励，根据目标速度矢量与飞机当前速度矢量的夹角给出的负奖励。例如：目标速度矢量与飞机当前速度矢量的夹角为a，奖励为：-(a / 180) ^ b，其中b为预设的常量。

    Args:
        RewardBase (_type_): _description_
    """
    def __init__(self, b: float = 1., log_history_reward: bool = True, my_logger: logging.Logger = None) -> None:
        self.b = b
        super().__init__(is_potential=False, log_history_reward=log_history_reward, my_logger=my_logger)
    
    def get_reward(self, state: Union[namedtuple, np.ndarray], **kwargs) -> float:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_mu" in kwargs, "参数中需要包括goal_mu，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        # assert "next_ve" in kwargs, "调用前，需要传入next_ve"
        # assert "next_vn" in kwargs, "调用前，需要传入next_vn"
        # assert "next_vh" in kwargs, "调用前，需要传入next_vh"
        assert "next_state" in kwargs, "调用前，需要把namedtuple类型的next_state放入函数的参数中"
        
        next_state = kwargs["next_state"]

        # plane_current_velocity_vector = [
        #     kwargs["next_ve"], 
        #     kwargs["next_vn"], 
        #     kwargs["next_vh"]
        # ]
        plane_current_velocity_vector = [
            next_state.v * np.cos(np.deg2rad(next_state.mu)) * np.sin(np.deg2rad(next_state.chi)), 
            next_state.v * np.cos(np.deg2rad(next_state.mu)) * np.cos(np.deg2rad(next_state.chi)), 
            next_state.v * np.sin(np.deg2rad(next_state.mu))
        ]
        target_velocity_vector = [
            kwargs["goal_v"] * np.cos(np.deg2rad(kwargs["goal_mu"])) * np.sin(np.deg2rad(kwargs["goal_chi"])), 
            kwargs["goal_v"] * np.cos(np.deg2rad(kwargs["goal_mu"])) * np.cos(np.deg2rad(kwargs["goal_chi"])),
            kwargs["goal_v"] * np.sin(np.deg2rad(kwargs["goal_mu"])),
        ]

        angle = angle_of_2_3d_vectors(plane_current_velocity_vector, target_velocity_vector)

        return -np.power(angle / 180., self.b)

    def reset(self):
        super().reset()


from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys
import numpy as np

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase
from utils.geometry_utils import angle_of_2_3d_vectors


class ContinuouselyMoveAwayTermination2(TerminationBase):
    """持续远离目标：连续2秒，当前速度矢量和目标速度矢量之间的误差增大
    """
    def __init__(self,
        time_window: float=2.,
        ignore_velocity_vector_angle_error: float=1.,
        termination_reward: float=-1.,
        is_termination_reward_based_on_steps_left: bool=False,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        """_summary_

        Args:
            step_frequence (int, optional): _description_. Defaults to 100.
            time_window (float, optional): _description_. Defaults to 2..
            termination_reward (float, optional): _description_. Defaults to -1..
            ignore_velocity_vector_angle_error (float, optional): 速度矢量误差大于这个值的时候，才考虑持续远离条件. Defaults to 1..
            my_logger (logging.Logger, optional): _description_. Defaults to None.
        """
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=is_termination_reward_based_on_steps_left,
            env_config=env_config,
            my_logger=my_logger
        )

        self.time_window = time_window
        self.ignore_velocity_vector_angle_error = ignore_velocity_vector_angle_error

        # 例子：
        # velocity_vector_errors:                          2, 3, 5, 3, 6, 7
        # velocity_vector_continuously_increasing_num 计数：0, 1, 2, 0, 1, 2
        self.prev_velocity_vector_errors = 180.  # 上一拍的速度矢量误差
        self.velocity_vector_continuously_increasing_num = -1  # 到当前step为止，连续增大的速度矢量误差的数量

    
    def _get_termination(self, state: namedtuple, goal_v: float, goal_mu: float, goal_chi: float, ve: float, vn: float, vh: float):
        
        plane_current_velocity_vector = [
            ve, 
            vn, 
            vh
        ]

        target_velocity_vector = [
            goal_v * np.cos(np.deg2rad(goal_mu)) * np.sin(np.deg2rad(goal_chi)), 
            goal_v * np.cos(np.deg2rad(goal_mu)) * np.cos(np.deg2rad(goal_chi)),
            goal_v * np.sin(np.deg2rad(goal_mu)),
        ]

        cur_velocity_vector_error = angle_of_2_3d_vectors(plane_current_velocity_vector, target_velocity_vector)
        # print('velocity angle: ', cur_velocity_vector_error)
        
        if cur_velocity_vector_error <= self.ignore_velocity_vector_angle_error:
            self.velocity_vector_continuously_increasing_num = 0
        else:
            self.velocity_vector_continuously_increasing_num = 0 if cur_velocity_vector_error <= self.prev_velocity_vector_errors else self.velocity_vector_continuously_increasing_num + 1

        self.prev_velocity_vector_errors = cur_velocity_vector_error

        if self.velocity_vector_continuously_increasing_num >= self.time_window_step_length:

            if self.logger is not None:
                self.logger.info(f"速度矢量误差时持续增大{self.time_window_step_length}拍！！！")
            
            terminated, truncated = True, False
        else:
            terminated, truncated = False, False

        return terminated, truncated
    
    def get_termination(self, state: namedtuple, **kwargs) -> Tuple[bool, bool]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_mu" in kwargs, "参数中需要包括goal_mu，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        # assert "ve" in kwargs, "参数中需要包括ve，再调用get_termination方法"
        # assert "vn" in kwargs, "参数中需要包括vn，再调用get_termination方法"
        # assert "vh" in kwargs, "参数中需要包括vh，再调用get_termination方法"

        return self._get_termination(
            state=state, 
            goal_v=kwargs["goal_v"],
            goal_mu=kwargs["goal_mu"],
            goal_chi=kwargs["goal_chi"],
            ve=state.v * np.cos(np.deg2rad(state.mu)) * np.sin(np.deg2rad(state.chi)), 
            vn=state.v * np.cos(np.deg2rad(state.mu)) * np.cos(np.deg2rad(state.chi)), 
            vh=state.v * np.sin(np.deg2rad(state.mu))
        )
    
    def get_termination_and_reward(self, state: namedtuple, **kwargs) -> Tuple[bool, bool, float]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_mu" in kwargs, "参数中需要包括goal_mu，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        # assert "ve" in kwargs, "参数中需要包括ve，再调用get_termination方法"
        # assert "vn" in kwargs, "参数中需要包括vn，再调用get_termination方法"
        # assert "vh" in kwargs, "参数中需要包括vh，再调用get_termination方法"
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"

        terminated, truncated = self._get_termination(
            state=state, 
            goal_v=kwargs["goal_v"],
            goal_mu=kwargs["goal_mu"],
            goal_chi=kwargs["goal_chi"],
            ve=state.v * np.cos(np.deg2rad(state.mu)) * np.sin(np.deg2rad(state.chi)), 
            vn=state.v * np.cos(np.deg2rad(state.mu)) * np.cos(np.deg2rad(state.chi)), 
            vh=state.v * np.sin(np.deg2rad(state.mu))
        )
        # reward = self.termination_reward if terminated else 0.

        return terminated, truncated, self.get_termination_penalty(terminated=terminated, steps_cnt=kwargs["step_cnt"])

    def reset(self):
        self.prev_velocity_vector_errors = 180.
        self.velocity_vector_continuously_increasing_num = -1

    @property
    def time_window_step_length(self) -> int:
        """判断使用的时间窗口step数

        Returns:
            _type_: _description_
        """
        return round(self.time_window * self.step_frequence)

    def __str__(self) -> str:
        return "continuousely_move_away_termination_based_on_velocity_vector_error"


import gymnasium as gym
from gymnasium import spaces
from gymnasium.utils import seeding
import numpy as np
import pandas as pd
from collections import namedtuple
from typing import List
import logging
from pathlib import Path
import sys
from copy import deepcopy
import gc
import time
from typing import Union, Dict

PROJECT_ROOT_DIR = Path(__file__).parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from planes.f16_plane import F16Plane
from tasks.velocity_vector_control_task import VelocityVectorControlTask
from utils.load_config import load_config
from utils.dict_utils import update_nested_dict
from terminations.reach_target_termination import ReachTargetTermination
from terminations.reach_target_termination2 import ReachTargetTermination2
from terminations.reach_target_termination_single_step import ReachTargetTerminationSingleStep


class FlyCraftEnv(gym.Env):

    def __init__(self, 
        custom_config: dict={},
        config_file: Union[Path, str]=Path(PROJECT_ROOT_DIR / "configs" / "NMR.json"),
    ) -> None:
        # define spaces
        state_mins = VelocityVectorControlTask.get_state_lower_bounds()
        state_maxs = VelocityVectorControlTask.get_state_higher_bounds()
        goal_mins = VelocityVectorControlTask.get_goal_lower_bounds()
        goal_maxs = VelocityVectorControlTask.get_goal_higher_bounds()
        self.observation_space = spaces.Dict(
            dict(
                observation = spaces.Box(low=np.array(state_mins, dtype=np.float32), high=np.array(state_maxs, dtype=np.float32)),  # phi, theta, psi, v, mu, chi, p, h
                desired_goal = spaces.Box(low=np.array(goal_mins, dtype=np.float32), high=np.array(goal_maxs, dtype=np.float32)),
                achieved_goal = spaces.Box(low=np.array(goal_mins, dtype=np.float32), high=np.array(goal_maxs, dtype=np.float32)),
            )
        )
        
        print(f"load config from: {config_file}")

        self.env_config: dict = load_config(config_file)
        update_nested_dict(self.env_config, custom_config)

        self.plane: F16Plane = F16Plane(env_config=self.env_config)
        self.task: VelocityVectorControlTask = VelocityVectorControlTask(
            plane=self.plane,
            np_random=self.np_random,
            env_config=self.env_config
        )
        
        action_mins = F16Plane.get_action_lower_bounds(self.plane.control_mode)
        action_maxs = F16Plane.get_action_higher_bounds(self.plane.control_mode)
        self.action_space = spaces.Box(low=np.array(action_mins, dtype=np.float32), high=np.array(action_maxs, dtype=np.float32))  # (p, nz, pla) if control_mode == "guidance_law_mode" else (ail, ele, rud, pla)
        
        # log data
        self.step_cnt = 0
        self.plane_state_dict_list: List[Dict] = []
        self.plane_state_namedtuple_list: List[namedtuple] = []
        self.action_list = []
        
        # use for debug
        self.debug_mode: bool = self.env_config.get("debug_mode")
        self.flag_str: str = self.env_config.get("flag_str")

        # used by HerReplayBuffer
        self.compute_reward = self.task.compute_reward
    
    def _get_obs(self) -> Dict[str, np.ndarray]:
        return {
            "observation": np.array(VelocityVectorControlTask.convert_dict_to_state_vars(self.plane_state_dict_list[-1]), dtype=np.float32),
            "achieved_goal": self.task.get_achieved_goal().astype(np.float32),
            "desired_goal": self.task.get_goal().astype(np.float32),
        }

    def reset(self, *, seed = None, options = None):
        """

        Args:
            seed (_type_, optional): _description_. Defaults to None.
            options (_type_, optional): _description_. Defaults to None.
        
        Returns:
            tuple[ObsType, dict[str, Any]]: obs, info
        """
        super().reset(seed=seed, options=options)
        if seed != None:
            self.task.np_random = self.np_random
            self.task.goal_sampler.np_random = self.np_random

        plane_state_dict = self.plane.reset()
        self.task.reset()

        self.step_cnt = 0
        self.plane_state_dict_list = [plane_state_dict]
        self.plane_state_namedtuple_list = [VelocityVectorControlTask.convert_dict_to_state_vars(plane_state_dict)]
        self.action_list = []

        info = {}
        info["plane_state"] = plane_state_dict

        return self._get_obs(), info

    def close(self):
        self.plane.close()

    def step(self, action):
        """_summary_

        Args:
            action (_type_): guidance_law_mode模式下action是制导律输出的[p, nz, pla]，end_to_end_mode模式下action是[ail, ele, rud, pla]

        Returns:
            tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]: next_obs, reward, terminated, truncated, info

            info = {
                "plane_state": dict,
                "plane_next_state": dict,
                "rewards": float,
                "termination": bool,
                "step": int,
                "is_success": bool,
            }
        """
        self.step_cnt += 1
        self.action_list.append(action)

        plane_state_dict = self.plane.step(action)
        self.plane_state_dict_list.append(deepcopy(plane_state_dict))
        plane_state_namedtuple = VelocityVectorControlTask.convert_dict_to_state_vars(plane_state_dict)
        self.plane_state_namedtuple_list.append(plane_state_namedtuple)
        
        # check obs for NaN!!!!!!!!!
        if np.any(np.isnan(plane_state_namedtuple)):
            tmp = []
            for item_obs, item_act in zip(self.plane_state_namedtuple_list, self.action_list):
                if self.debug_mode:
                    print(item_obs, item_act)
                tmp.append([*item_obs, *item_act])
            
            if self.plane.control_mode_guidance_law:
                tmp_pd = pd.DataFrame(data=tmp, columns=['s_phi', 's_theta', 's_psi', 's_v', 's_mu', 's_chi', 's_p', 's_h', 'target_v', 'target_mu', 'target_chi','a_p', 'a_nz', 'a_pla'])
            else:
                tmp_pd = pd.DataFrame(data=tmp, columns=['s_phi', 's_theta', 's_psi', 's_v', 's_mu', 's_chi', 's_p', 's_h', 'target_v', 'target_mu', 'target_chi','a_ail', 'a_ele', 'a_rud', 'a_pla'])
            tmp_pd.to_csv(str((PROJECT_ROOT_DIR / "my_logs" / 'nan_states.csv').absolute()), index=False)

        terminated, truncated, reward, info = False, False, 0., {}
        info["step"] = self.step_cnt
        # tmp_goal = self.task.get_goal()
        # info["target_v"] = tmp_goal[0]
        # info["target_mu"] = tmp_goal[1]
        # info["target_chi"] = tmp_goal[2]
        info['is_success'] = False  # https://stable-baselines3.readthedocs.io/en/feat-gymnasium-support/common/logger.html#eval 评估中success_rate要求这个字段

        desired_goal = self.task.get_goal()

        # judge termination
        for t_func in self.task.termination_funcs:
            terminated, truncated, reward = t_func.get_termination_and_reward(
                state=self.plane_state_namedtuple_list[-2], 
                next_state=self.plane_state_namedtuple_list[-1],  # ContinuouselyRollTermination,
                step_cnt=self.step_cnt,  # TimeoutTermination,
                state_list=self.plane_state_namedtuple_list,  # ReachTargetTermination, 
                nz=self.plane_state_dict_list[-1]["nz"],  # NegativeOverloadAndBigPhiTermination,
                goal_v=desired_goal[0],
                goal_mu=desired_goal[1],
                goal_chi=desired_goal[2]
            )
            if terminated:
                info["termination"] = str(t_func)
                
                if isinstance(t_func, ReachTargetTermination) or isinstance(t_func, ReachTargetTermination2) or isinstance(t_func, ReachTargetTerminationSingleStep):
                    info['is_success'] = True

                if self.debug_mode:
                    if isinstance(t_func, ReachTargetTermination) or isinstance(t_func, ReachTargetTermination2) or isinstance(t_func, ReachTargetTerminationSingleStep):
                        # reach target, print with green
                        print(f"print, {self.flag_str}, \033[32m{str(t_func)}。\033[0m steps: {self.step_cnt}。target: ({desired_goal[0]:.2f}, {desired_goal[1]:.2f}, {desired_goal[2]:.2f})。achieved target: ({plane_state_dict['v']:.2f}, {plane_state_dict['mu']:.2f}, {plane_state_dict['chi']:.2f})。expert steps: {self.task.goal_sampler.goal_expert_length}。")
                    else:
                        # the other terminations, print with red
                        print(f"print, {self.flag_str}, \033[31m{str(t_func)}。\033[0m steps: {self.step_cnt}。target: ({desired_goal[0]:.2f}, {desired_goal[1]:.2f}, {desired_goal[2]:.2f})。achieved target: ({plane_state_dict['v']:.2f}, {plane_state_dict['mu']:.2f}, {plane_state_dict['chi']:.2f})。expert steps: {self.task.goal_sampler.goal_expert_length}。")

                if "rewards" not in info:
                    info["rewards"] = {}
                info["rewards"][str(t_func)] = reward
                break
        
        # compute reward
        for r_func in self.task.reward_funcs:
            tmp_reward = r_func.get_reward(
                state=self.plane_state_namedtuple_list[-2], 
                next_state=self.plane_state_namedtuple_list[-1],
                done=terminated,
                goal_v=desired_goal[0],
                goal_mu=desired_goal[1],
                goal_chi=desired_goal[2],
            )
            if "rewards" not in info:
                info["rewards"] = {}
            info["rewards"][str(r_func)] = tmp_reward
            reward += tmp_reward / len(self.task.reward_funcs)
        
        if self.plane.control_mode_guidance_law:
            info["action"] = {
                "p": action[0],
                "nz": action[1],
                "pla": action[2],
                "rud": 0.
            }
        else:
            info["action"] = {
                "ail": action[0],
                "ele": action[1],
                "rud": action[2],
                "pla": action[3]
            }
        info["desired_goal"] = [*desired_goal]
        info["plane_state"] = deepcopy(self.plane_state_dict_list[-2])
        info["plane_next_state"] = deepcopy(self.plane_state_dict_list[-1])

        return self._get_obs(), reward, terminated, truncated, info



from typing import Any, Dict, List, Union
import numpy as np
from collections import namedtuple
import logging
from gymnasium.utils import seeding

from planes.f16_plane import F16Plane
from tasks.task_base import Task
from tasks.goal_samplers.goal_sampler_for_velocity_vector_control import GoalSampler

from rewards.reward_base import RewardBase
from rewards.dense_reward_based_on_angle_and_velocity import DenseRewardBasedOnAngleAndVelocity
from rewards.sparse_reward import SparseReward

from terminations.termination_base import TerminationBase
from terminations.reach_target_termination2 import ReachTargetTermination2
from terminations.reach_target_termination_single_step import ReachTargetTerminationSingleStep
from terminations.crash_termination import CrashTermination
from terminations.extreme_state_termination import ExtremeStateTermination
from terminations.timeout_termination import TimeoutTermination
from terminations.continuousely_move_away_termination import ContinuouselyMoveAwayTermination
from terminations.continuousely_roll_termination import ContinuouselyRollTermination
from terminations.negative_overload_and_big_phi_termination import NegativeOverloadAndBigPhiTermination

class VelocityVectorControlTask(Task):

    def __init__(
        self, 
        plane: F16Plane,
        env_config: dict,
        np_random: np.random.Generator=None,
        my_logger: logging.Logger=None
    ) -> None:
        super().__init__(plane)
        self.config = {
            "flag_str": "Train",
            "debug_mode": True,
            "task": {
                "h0": 5000,
                "v0": 200,
                "step_frequence": 10,
                "max_simulate_time": 40,
                "gamma": 0.995
            },
            "goal": {
                "use_fixed_goal": False,
                "goal_v": 0, 
                "goal_mu": 0, 
                "goal_chi": 0,
                "sample_random": True,
                "v_min": 100., 
                "v_max": 300.,
                "mu_min": -90.,
                "mu_max": 90.,
                "chi_min": -180.,
                "chi_max": 180.,
                "available_goals_file": "res.csv",
                "sample_reachable_goal": False,
                "sample_goal_noise_std": [5, 0.5, 0.5],  # noise std for [v, mu, chi]
            },
            "rewards": {
                "dense": {
                    "use": True,
                    "b": 0.5
                },
            },
            "terminations": {
                "RT": {
                    "use": False, 
                    "integral_time_length": 1,  # 在连续多长的时间窗口内满足条件
                    "v_threshold": 10,  # 触发RT的最大v误差
                    "angle_threshold": 3,  # 触发RT的最大角度误差
                    "termination_reward": 0.0,  # 触发RT的奖励
                },
                "RT_SINGLE_STEP": {
                    "use": True,
                    "v_threshold": 10,
                    "angle_threshold": 3,
                    "termination_reward": 0.0
                },
                "C": {
                    "use": True,
                    "h0": 0,
                    "is_termination_reward_based_on_steps_left": True,  # 是否根据剩余步长计算惩罚
                    "termination_reward": -1,  # 不跟据剩余步长计算惩罚时使用的立即惩罚大小
                },
                "ES": {
                    "use": True,
                    "v_max": 400,
                    "p_max": 300,
                    "is_termination_reward_based_on_steps_left": True,  # 是否根据剩余步长计算惩罚
                    "termination_reward": -1,  # 不跟据剩余步长计算惩罚时使用的立即惩罚大小
                },
                "T": {
                    "use": True,
                    "termination_reward": -1,
                },
                "CMA": {
                    "use": True,
                    "time_window": 2,  # 在这个时间窗口内持续远离目标，就触发该终止条件
                    "ignore_mu_error": 1,  # 当mu的误差小于ignore_mu_error时，不再考虑这个终止条件
                    "ignore_chi_error": 1,  # 当chi的误差小于ignore_chi_error时，不再考虑这个终止条件
                    "is_termination_reward_based_on_steps_left": True,  # 是否根据剩余步长计算惩罚
                    "termination_reward": -1,  # 不跟据剩余步长计算惩罚时使用的立即惩罚大小
                },
                "CR": {
                    "use": True,
                    "continuousely_roll_threshold": 720,
                    "is_termination_reward_based_on_steps_left": True,  # 是否根据剩余步长计算惩罚
                    "termination_reward": -1,  # 不跟据剩余步长计算惩罚时使用的立即惩罚大小
                },
                "NOBR": {
                    "use": True,
                    "time_window": 2,  # 负过载且大幅度滚转：nz<0且phi>60(deg)连续超过time_window秒，就触发该终止条件
                    "negative_overload_threshold": 0,
                    "big_phi_threshold": 60,
                    "is_termination_reward_based_on_steps_left": True,  # 是否根据剩余步长计算惩罚
                    "termination_reward": -1,  # 不跟据剩余步长计算惩罚时使用的立即惩罚大小
                }
            }
        }
        self.config.update(env_config)

        self.reward_funcs: List[RewardBase] = []
        self._prep_reward_funcs()

        self.termination_funcs: List[TerminationBase] = []
        self._prep_termination_funcs()

        if np_random != None:
            self.np_random: np.random.Generator = np_random
        else:
            self.np_random, _ = seeding.np_random(np.random.randint(0, 1_000_000))
        
        self.goal_sampler: GoalSampler = GoalSampler(
            np_random=self.np_random, 
            env_config=self.config
        )

        self.logger: logging.Logger = my_logger

    def _prep_reward_funcs(self):
        self.reward_funcs = []
        for rwd in self.config["rewards"]:
            tmp_cfg = self.config["rewards"][rwd]
            if tmp_cfg["use"]:
                if rwd == "dense":
                    self.reward_funcs.append(
                        DenseRewardBasedOnAngleAndVelocity(
                            b=tmp_cfg.get("b", 0.5),
                            angle_scale=tmp_cfg.get("angle_scale", 180),
                            velocity_scale=tmp_cfg.get("velocity_scale", 100),
                            angle_weight=tmp_cfg.get("angle_weight", 0.5),
                        )
                    )
                elif rwd == "sparse":
                    self.reward_funcs.append(
                        SparseReward(
                            reward_constant=tmp_cfg.get("reward_constant", 0.0),
                        )
                    )
                # TODO: other rewards

    def _prep_termination_funcs(self):
        # TODO: 如何设置优先级？？？
        self.termination_funcs = []

        for tmnt in self.config["terminations"]:
            tmp_cfg = self.config["terminations"][tmnt]
            if tmp_cfg["use"]:
                if tmnt == "RT":
                    self.termination_funcs.append(
                        ReachTargetTermination2(
                            integral_time_length=tmp_cfg["integral_time_length"],
                            v_threshold=tmp_cfg["v_threshold"],
                            angle_threshold=tmp_cfg["angle_threshold"],
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                            # my_logger=self.logger
                        )
                    )
                elif tmnt == "RT_SINGLE_STEP":
                    self.termination_funcs.append(
                        ReachTargetTerminationSingleStep(
                            v_threshold=tmp_cfg["v_threshold"],
                            angle_threshold=tmp_cfg["angle_threshold"],
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                        )
                    )
                elif tmnt == "C":
                    self.termination_funcs.append(
                        CrashTermination(
                            h0=tmp_cfg["h0"],
                            is_termination_reward_based_on_steps_left=tmp_cfg["is_termination_reward_based_on_steps_left"],
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                            # my_logger=self.logger
                        )
                    )
                elif tmnt == "ES":
                    self.termination_funcs.append(
                        ExtremeStateTermination(
                            v_max=tmp_cfg["v_max"],
                            p_max=tmp_cfg["p_max"],
                            is_termination_reward_based_on_steps_left=tmp_cfg["is_termination_reward_based_on_steps_left"],
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                            # my_logger=self.logger
                        )
                    )
                elif tmnt == "T":
                    self.termination_funcs.append(
                        TimeoutTermination(
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                            # my_logger=self.logger
                        )
                    )
                elif tmnt == "CMA":
                    self.termination_funcs.append(
                        ContinuouselyMoveAwayTermination(
                            time_window=tmp_cfg["time_window"],
                            ignore_mu_error=tmp_cfg["ignore_mu_error"],
                            ignore_chi_error=tmp_cfg["ignore_chi_error"],
                            is_termination_reward_based_on_steps_left=tmp_cfg["is_termination_reward_based_on_steps_left"],
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                            # my_logger=self.logger
                        )
                    )
                elif tmnt == "CR":
                    self.termination_funcs.append(
                        ContinuouselyRollTermination(
                            continuousely_roll_threshold=tmp_cfg["continuousely_roll_threshold"],
                            is_termination_reward_based_on_steps_left=tmp_cfg["is_termination_reward_based_on_steps_left"],
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                            # my_logger=self.logger
                        )
                    )
                elif tmnt == "NOBR":
                    self.termination_funcs.append(
                        NegativeOverloadAndBigPhiTermination(
                            time_window=tmp_cfg["time_window"],
                            negative_overload_threshold=tmp_cfg["negative_overload_threshold"],
                            big_phi_threshold=tmp_cfg["big_phi_threshold"],
                            is_termination_reward_based_on_steps_left=tmp_cfg["is_termination_reward_based_on_steps_left"],
                            termination_reward=tmp_cfg["termination_reward"],
                            env_config=self.config,
                            # my_logger=self.logger
                        )
                    )

    def get_obs(self) -> np.ndarray:
        current_state_dict = self.plane.current_obs
        return np.array([
            current_state_dict['phi'],
            current_state_dict['theta'],
            current_state_dict['psi'],
            current_state_dict['v'],
            current_state_dict['mu'],
            current_state_dict['chi'],
            current_state_dict['p'],
            current_state_dict['h']
        ])

    def get_achieved_goal(self) -> np.ndarray:
        current_state_dict = self.plane.current_obs
        return np.array([
            current_state_dict['v'],
            current_state_dict['mu'],
            current_state_dict['chi'],
        ])

    def reset(self) -> None:
        goal_dict = self.goal_sampler.sample_goal()
        self.goal = np.array([goal_dict["v"], goal_dict["mu"], goal_dict["chi"]])
        
        for t_func in self.termination_funcs:
            t_func.reset()
        
        for r_func in self.reward_funcs:
            r_func.reset()
    
    def is_success(
        self, 
        achieved_goal: np.ndarray, 
        desired_goal: np.ndarray, 
        info: Dict[str, Any] = {}
    ) -> np.ndarray:
        """use only ReachTargetTerminationSingleStep to judge whether achieving desired goal for off-policy RL algorithms.

        Args:
            achieved_goal (np.ndarray): _description_
            desired_goal (np.ndarray): _description_
            info (Dict[str, Any], optional): _description_. Defaults to {}.

        Returns:
            np.ndarray: _description_
        """

        # pick out the reach target termination
        reach_target_termination_func = None
        for t in self.termination_funcs:
            if isinstance(t, ReachTargetTerminationSingleStep):
                reach_target_termination_func = t
                break
        
        if reach_target_termination_func == None:
            raise ValueError("VelocityVectorControlTask: when using off-policy algorithms, must use the termination condition: ReachTargetTerminationSingleStep!!!")

        # make tmp_achieved_goal and tmp_desired_goal be of shape (batch, goal_dim)
        if len(achieved_goal.shape) == 1:
            tmp_achieved_goal = achieved_goal.reshape((1, -1))
            tmp_desired_goal = desired_goal.reshape((1, -1))
        elif len(achieved_goal.shape) == 2:
            tmp_achieved_goal = achieved_goal
            tmp_desired_goal = desired_goal
        else:
            raise ValueError("VelocityVectorControlTask: the shape of achieved goal mush be 1-D or 2-D!")

        terminated_arr = []
        for tmp_a, tmp_d in zip(tmp_achieved_goal, tmp_desired_goal):
            state_var = VelocityVectorControlTask.get_state_vars()
            cur_state_namedtuple = state_var(phi=0, theta=0, psi=0, v=tmp_a[0], mu=tmp_a[1], chi=tmp_a[2], p=0, h=0)

            ternimated, truncated = reach_target_termination_func.get_termination(
                state=cur_state_namedtuple,
                goal_v=tmp_d[0],
                goal_mu=tmp_d[1],
                goal_chi=tmp_d[2]
            )

            terminated_arr.append(ternimated)
        
        if len(achieved_goal.shape) == 1:
            return terminated_arr[0]
        else:
            return np.array(terminated_arr, dtype=bool)

    def compute_reward(
        self, 
        achieved_goal: np.ndarray, 
        desired_goal: np.ndarray, 
        info: Union[Dict[str, Any], List[Dict]] = {}
    ) -> np.ndarray:
        
        if len(achieved_goal.shape) == 1:
            tmp_achieved_goals = achieved_goal.reshape((1, -1))
            tmp_desired_goals = desired_goal.reshape((1, -1))
            tmp_infos = [info]
        elif len(achieved_goal.shape) == 2:
            tmp_achieved_goals = achieved_goal
            tmp_desired_goals = desired_goal
            tmp_infos = info
        else:
            raise ValueError("VelocityVectorControlTask: the shape of achieved goal mush be 1-D or 2-D!")
        
        # compute reward: base on self.reward_funcs
        reward_arr = []
        state_var = VelocityVectorControlTask.get_state_vars()

        for tmp_a, tmp_d, tmp_info in zip(tmp_achieved_goals, tmp_desired_goals, tmp_infos):
            # 使用self.reward_funcs中的所有奖励函数计算reward
            reward = 0.
            for r_func in self.reward_funcs:
                tmp_reward = r_func.get_reward(
                    state=state_var(0., 0., 0., 0., 0., 0., 0., 0.),
                    goal_v=tmp_d[0],
                    goal_mu=tmp_d[1],
                    goal_chi=tmp_d[2],
                    next_state=state_var(0., 0., 0., tmp_a[0], tmp_a[1], tmp_a[2], 0., 0.)
                )
                reward += tmp_reward / len(self.reward_funcs)
            reward_arr.append(reward)
        
        if len(achieved_goal.shape) == 1:
            return reward_arr[0]
        else:
            return np.array(reward_arr, dtype=np.float32)
        
    @staticmethod
    def get_state_vars():
        """学习器使用的观测

        Returns:
            _type_: _description_
        """
        return namedtuple("state_vars", ["phi", "theta", "psi", "v", "mu", "chi", "p", "h"])
    
    @staticmethod
    def get_goal_vars():

        return namedtuple("goal_vars", ["v", "mu", "chi"])
    
    @staticmethod
    def convert_dict_to_state_vars(state_dict:dict) -> namedtuple:
        """将仿真器返回的字典类型观测转换为环境定义的观测(VelocityVectorControlTask.get_state_vars()定义的namedtuple)

        Args:
            state_dict (dict): 键包括：lef, npos, epos, h, alpha, beta, phi, theta, psi, p, q, r, v, vn, ve, vh, nx, ny, nz, ele, ail, rud, thrust, lon, lat, mu, chi

        Returns:
            namedtuple: _description_
        """

        state_vars_type = VelocityVectorControlTask.get_state_vars()
        return state_vars_type(
            phi=state_dict['phi'], theta=state_dict['theta'], psi=state_dict['psi'], 
            v=state_dict['v'], mu=state_dict['mu'], chi=state_dict['chi'],
            p=state_dict['p'], h=state_dict['h']
        )

    @staticmethod
    def get_state_lower_bounds():
        """返回观测的下限

        Returns:
            _type_: _description_
        """
        state_vars_type = VelocityVectorControlTask.get_state_vars()
        return state_vars_type(phi=-180., theta=-90., psi=-180., v=0., mu=-90., chi=-180., p=-300., h=0.)

    @staticmethod
    def get_state_higher_bounds():
        """返回观测的上限

        Returns:
            _type_: _description_
        """
        state_vars_type = VelocityVectorControlTask.get_state_vars()
        return state_vars_type(phi=180., theta=90., psi=180., v=1000., mu=90., chi=180., p=300., h=20000.)

    @staticmethod
    def get_goal_lower_bounds():
        goal_vars_type = VelocityVectorControlTask.get_goal_vars()
        return goal_vars_type(v=0., mu=-90., chi=-180.)
    
    @staticmethod
    def get_goal_higher_bounds():
        goal_vars_type = VelocityVectorControlTask.get_goal_vars()
        return goal_vars_type(v=1000., mu=90., chi=180.)
    


from abc import ABC, abstractmethod
import numpy as np
from collections import namedtuple
from typing import Union, Callable
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from rewards.reward_base import RewardBase


class PonentialRewardBasedOnVelocity(RewardBase, ABC):
    """基于速度矢量大小的shaping reward

    Args:
        RewardBase (_type_): _description_
        ABC (_type_): _description_
    """

    def __init__(self, b: float=1., scale: float=100., gamma: float=0.99, log_history_reward: bool = True, my_logger: logging.Logger = None) -> None:
        self.b = b
        self.gamma = gamma
        self.scale = scale
        super().__init__(is_potential=False, log_history_reward=log_history_reward, my_logger=my_logger)
    
    def get_reward(self, state: Union[namedtuple, np.ndarray], **kwargs) -> float:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "next_state" in kwargs, "调用前，需要把namedtuple类型的next_state放入函数的参数中"
        assert "done" in kwargs, ""

        next_state = kwargs["next_state"]
        done = kwargs["done"]

        reward = self.gamma * (0. if done else self.phi(next_state, kwargs["goal_v"])) - self.phi(state, kwargs["goal_v"])
        
        return self._process(new_reward=reward)

    def phi(self, state: namedtuple, goal_v: float):
        delta_v = np.abs(goal_v - state.v)
        return -np.power(delta_v / self.scale, self.b)

    def reset(self):
        super().reset()



import os

from setuptools import find_packages, setup

with open(os.path.join("flycraft", "version.txt"), "r") as file_handler:
    __version__ = file_handler.read().strip()

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="flycraft",
    description="A fixed-wing UAV environment based on gymnasium.",
    author="Xudong Gong",
    author_email="gongxudong_cs@aliyun.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gongxudong/fly-craft",
    packages=find_packages(),
    include_package_data=True,
    # package_data={"flycraft": ["version.txt"]},
    version=__version__,
    install_requires=["gymnasium>=0.26", "numpy", "pandas"],
    extras_require={
        "develop": ["pytest-cov", "black", "isort", "pytype", "sphinx", "sphinx-rtd-theme"],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys
import numpy as np

PROJECT_ROOT_DIR = Path(__file__).parent.parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase


class ReachTargetTerminationSingleStep(TerminationBase):

    def __init__(self,
        mu_tolerance: float=1.,
        v_threshold: float=10., 
        chi_threshold: float=1.,
        termination_reward: float=1.,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=False,
            env_config=env_config, 
            my_logger=my_logger
        )
        
        self.mu_tolerance = mu_tolerance
        self.v_threshold = v_threshold
        self.chi_threshold = chi_threshold

    def _get_termination(
        self, 
        v: float, mu: float, chi: float, 
        goal_v: float, goal_chi: float
    ) -> Tuple[bool, bool]:
        
        if np.abs(mu) < self.mu_tolerance:
            v_flag, chi_flag = False, False
            if abs(goal_v - v) < self.v_threshold:
                v_flag = True
            if abs(goal_chi - chi) < self.chi_threshold:
                chi_flag = True
            
            if v_flag and chi_flag:
                return True, False
            else:
                return False, False
        else:
            return False, False

    def get_termination(self, state, **kwargs) -> Tuple[bool, bool]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"

        return self._get_termination(
            v=state.v,
            mu=state.mu,
            chi=state.chi,
            goal_v=kwargs["goal_v"], 
            goal_chi=kwargs["goal_chi"],
        )

    def get_termination_and_reward(self, state, **kwargs) -> Tuple[bool, bool, float]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"

        terminated, truncated = self._get_termination(
            v=state.v,
            mu=state.mu,
            chi=state.chi,
            goal_v=kwargs["goal_v"], 
            goal_chi=kwargs["goal_chi"],
        )
        reward = self.termination_reward if terminated else 0.
        
        return terminated, truncated, reward

    def reset(self):
        pass

    def __str__(self) -> str:
        return "reach_target_termination_single_step_based_on_chi_and_velocity"


import pandas as pd
from pathlib import Path

PROJECT_ROOT_DIR = Path(__file__).parent.parent.parent

def csv2acmi(
    traj_df: pd.DataFrame, 
    plane_id: str="A0100",
    time_column_name: str="time",
    lon_column_name: str="Longitude", 
    lat_column_name: str="Latitude", 
    alt_column_name: str="Altitude", 
    roll_column_name: str="Roll", 
    pitch_column_name: str="Pitch", 
    yaw_column_name: str="Yaw",
    acmi_file_save_dir: Path=PROJECT_ROOT_DIR / "expert_traj" / "acmi",
    acmi_file_name: str="test.acmi",
):
    assert time_column_name in traj_df.columns, f"DataFrame中不包含{time_column_name}列！！！"
    assert lon_column_name in traj_df.columns, f"DataFrame中不包含{lon_column_name}列！！！"
    assert lat_column_name in traj_df.columns, f"DataFrame中不包含{lat_column_name}列！！！"
    assert alt_column_name in traj_df.columns, f"DataFrame中不包含{alt_column_name}列！！！"
    assert roll_column_name in traj_df.columns, f"DataFrame中不包含{roll_column_name}列！！！"
    assert pitch_column_name in traj_df.columns, f"DataFrame中不包含{pitch_column_name}列！！！"
    assert yaw_column_name in traj_df.columns, f"DataFrame中不包含{yaw_column_name}列！！！"

    acmi_file = acmi_file_save_dir / acmi_file_name

    with open(acmi_file, mode='w', encoding='utf-8-sig') as f:
            f.write("FileType=text/acmi/tacview\n")
            f.write("FileVersion=2.1\n")
            f.write("0,ReferenceTime=2023-05-01T00:00:00Z\n")

    with open(acmi_file, mode='a', encoding='utf-8-sig') as f:
        for index, row in traj_df.iterrows():
            f.write(f"#{row[time_column_name]:.2f}\n")
            out_str = f"{plane_id},T={row[lon_column_name]}|{row[lat_column_name]}|{row[alt_column_name]}|{row[roll_column_name]}|{row[pitch_column_name]}|{row[yaw_column_name]},Name=F16,Color=Red\n"
            f.write(out_str)


from typing import Tuple, List
from collections import namedtuple
import logging
from pathlib import Path
import sys

PROJECT_ROOT_DIR = Path(__file__).parent.parent
if str(PROJECT_ROOT_DIR.absolute()) not in sys.path:
    sys.path.append(str(PROJECT_ROOT_DIR.absolute()))

from terminations.termination_base import TerminationBase


class ContinuouselyMoveAwayTermination(TerminationBase):
    """持续远离目标：连续2秒，当前状态和目标状态之间的角度差增大。（当mu的误差小于ignore_mu_error时，不再考虑这个终止条件；chi同理）
    """
    def __init__(self,
        time_window: float=2.,
        ignore_mu_error: float=1.,
        ignore_chi_error: float=1.,
        termination_reward: float=-1.,
        is_termination_reward_based_on_steps_left: bool=False,
        env_config: dict=None,
        my_logger: logging.Logger=None
    ) -> None:
        super().__init__(
            termination_reward=termination_reward, 
            is_termination_reward_based_on_steps_left=is_termination_reward_based_on_steps_left,
            env_config=env_config,
            my_logger=my_logger
        )

        self.time_window = time_window
        self.ignore_mu_error = ignore_mu_error
        self.ignore_chi_error = ignore_chi_error

        # 例子：
        # mu errors:                          2, 3, 5, 3, 6, 7
        # mu_continuously_increasing_num 计数：0, 1, 2, 0, 1, 2
        self.prev_mu_errors = -360.  # 上一拍的mu误差
        self.mu_continuously_increasing_num = -1  # 到当前step为止，连续增大的mu误差的数量
        self.prev_chi_errors = -360.  # 上一拍的chi误差
        self.chi_continuously_increasing_num = -1  # 到当前step为止，连续增大的chi误差的数量

    
    def _get_termination(self, state: namedtuple, goal_v: float, goal_chi: float):
        # mu的范围是[-90, 90]，不存在-90与90的跨越问题
        cur_mu_error = abs(0. - state.mu)

        # chi的范围是[-180, 180]，存在-180与180的跨越问题
        # 例如，从target_chi=-179, chi=179, 此时tmp_chi_error=358, 二者间的误差实际应该是360 - 358 = 2
        tmp_chi_error = abs(goal_chi - state.chi)
        cur_chi_error = min(tmp_chi_error, 360. - tmp_chi_error)

        # 当mu和chi的error很小时，不再考虑这个终止条件
        if cur_mu_error <= self.ignore_mu_error:
            self.mu_continuously_increasing_num = 0
        else:
            self.mu_continuously_increasing_num = 0 if cur_mu_error <= self.prev_mu_errors else self.mu_continuously_increasing_num + 1
        
        if cur_chi_error <= self.ignore_chi_error:
            self.chi_continuously_increasing_num = 0
        else:
            self.chi_continuously_increasing_num = 0 if cur_chi_error <= self.prev_chi_errors else self.chi_continuously_increasing_num + 1
        
        self.prev_mu_errors, self.prev_chi_errors = cur_mu_error, cur_chi_error

        if self.mu_continuously_increasing_num >= self.time_window_step_length and \
            self.chi_continuously_increasing_num >= self.time_window_step_length:

            if self.logger is not None:
                self.logger.info(f"mu和chi同时持续增大{self.time_window_step_length}拍！！！")
            
            terminated, truncated = True, False
        # 注意：mu和chi的误差，只有一个持续增大时，并不足以说明飞机持续远离目标！！！
        # 一个能够说明的例子：当目标v=130，mu=10，chi=145时，专家轨迹mu先减小到-16在逐渐增大到10，chi持续增大到145.
        # mu虽然持续远离了目标mu（原因是转弯掉高），但chi误差却减小了，飞机速度矢量与目标速度矢量靠近了！！！

        # elif self.mu_continuously_increasing_num >= self.time_window_step_length and \
        #     self.chi_continuously_increasing_num < self.time_window_step_length:

        #     if self.logger is not None:
        #         self.logger.info(f"mu持续增大{self.time_window_step_length}拍！！！")
            
        #     terminated, truncated = True, False
        # elif self.mu_continuously_increasing_num < self.time_window_step_length and \
        #     self.chi_continuously_increasing_num >= self.time_window_step_length:

        #     if self.logger is not None:
        #         self.logger.info(f"chi持续增大{self.time_window_step_length}拍！！！")
            
        #     terminated, truncated = True, False
        else:
            terminated, truncated = False, False

        return terminated, truncated
    
    def get_termination(self, state: namedtuple, **kwargs) -> Tuple[bool, bool]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"

        return self._get_termination(
            state=state, 
            goal_v=kwargs["goal_v"], 
            goal_chi=kwargs["goal_chi"]
        )
    
    def get_termination_and_reward(self, state: namedtuple, **kwargs) -> Tuple[bool, bool, float]:
        assert "goal_v" in kwargs, "参数中需要包括goal_v，再调用get_termination方法"
        assert "goal_chi" in kwargs, "参数中需要包括goal_chi，再调用get_termination方法"
        assert "step_cnt" in kwargs, "参数中需要包括step_cnt"

        terminated, truncated = self._get_termination(
            state=state,
            goal_v=kwargs["goal_v"], 
            goal_chi=kwargs["goal_chi"]
        )
        # reward = self.termination_reward if terminated else 0.

        return terminated, truncated, self.get_termination_penalty(terminated=terminated, steps_cnt=kwargs["step_cnt"])

    def reset(self):
        self.prev_mu_errors = -360.
        self.mu_continuously_increasing_num = -1
        self.prev_chi_errors = -360.
        self.chi_continuously_increasing_num = -1

    @property
    def time_window_step_length(self) -> int:
        """判断使用的时间窗口step数

        Returns:
            _type_: _description_
        """
        return round(self.time_window * self.step_frequence)

    def __str__(self) -> str:
        return "continuousely_move_away_termination_based_on_mu_error_and_chi_error"


#!/user/bin/python
# -*- coding: utf-8 -*-
'''LoggerOld is used to store texts into the indicated file on disk
Created by Weijia on 2017/06/01
Last modified by Weijia on 2018/08/04
'''
import os
loggerList = {}

class LoggerOld(object):
    def __init__(self, path, append = False):
        self.name = path
        if append:
            loggerList[self.name] = open(self.name ,'a')
        else:
            loggerList[self.name] = open(self.name ,'w')

    def write(self, str, end = "\n"):
        try:
            loggerList[self.name].write(str+end)
        except:
            writeLog(str, self.name)
            # print "logger "+self.name+" reinitialized"
    
    def continuousWrite(self, str):
        self.write(str,"")    
    
    def __del__(self):
        closeLog(self.name)

def hasLogger(name):
    return name in loggerList
    
def writeLog(str, name, end = "\n"):
    if hasLogger(name):
        loggerList[name].write(str+end)
    else:
        loggerList[name] = open(name ,'w')
        loggerList[name].write(str+end)
    loggerList[name].flush()

def closeLog(logName):
    if logName in loggerList:
        loggerList[logName].close()
        loggerList.pop(logName)